# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.results.result_record_entry_component import ResultRecordEntryComponent
from availsim4core.src.results.result_record_entry_phase import ResultRecordEntryPhase
from availsim4core.src.results.simulation_results import SimulationResults
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.record_generator import RecordGenerator
from availsim4core.src.timeline.record_phase import RecordPhase
from availsim4core.src.timeline.timeline import Timeline


class test_SimulationResults(unittest.TestCase):

    def test_update_with_des_results(self):
        basic_1 = Basic(1, "test_basic_1", 0, [], [], [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], [], [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)
        record_list = [RecordPhase(PhaseManager.NO_PHASE,
                                   0,
                                   RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT)
                       ]

        timeline = Timeline()
        timeline.record_list = record_list
        simulation_results = SimulationResults(10, 50)
        simulation_results.update_with_des_results({'number_of_time_iteration':5}, timeline)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].min, 5)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].max, 5)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].sum, 5)

        result_record_phase = ResultRecordEntryPhase(PhaseManager.NO_PHASE,RecordGenerator.DESCRIPTION_INIT)
        result_record_basic1 = ResultRecordEntryComponent(basic_1, Status.RUNNING, PhaseManager.NO_PHASE,RecordGenerator.DESCRIPTION_INIT)
        result_record_basic2 = ResultRecordEntryComponent(basic_2, Status.RUNNING, PhaseManager.NO_PHASE,RecordGenerator.DESCRIPTION_INIT)
        result_record_compound = ResultRecordEntryComponent(compound, Status.RUNNING, PhaseManager.NO_PHASE,RecordGenerator.DESCRIPTION_INIT)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_phase].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic1].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic2].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_compound].min, 1)

        self.assertEqual(simulation_results.duration_statistics[result_record_phase].min, 50)
        self.assertEqual(simulation_results.duration_statistics[result_record_basic1].min, 50)
        self.assertEqual(simulation_results.duration_statistics[result_record_basic2].min, 50)
        self.assertEqual(simulation_results.duration_statistics[result_record_compound].min, 50)

    def test_update_with_des_results_multiple_updates(self):
        phase_first = Phase("phase_test", DeterministicLaw(10), True)
        phase_second = Phase("phase_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)

        basic_1 = Basic(1, "test_basic_1", 0, [], [], [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], [], [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)
        record_list = [RecordPhase(phase_first,
                                   0,
                                   RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       0,
                                       phase_first,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       0,
                                       phase_first,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       0,
                                       phase_first,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordPhase(phase_second,
                                   10,
                                   PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       10,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       10,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       10,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordPhase(phase_first,
                                   20,
                                   PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       20,
                                       phase_first,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       20,
                                       phase_first,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       20,
                                       phase_first,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_1,
                                       Status.FAILED,
                                       21,
                                       phase_first,
                                       "Failure"),
                       RecordComponent(basic_1,
                                       Status.UNDER_REPAIR,
                                       21,
                                       phase_first,
                                       "Repair"),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       22,
                                       phase_first,
                                       "Repair"),
                       RecordPhase(phase_second,
                                   22,
                                   PhaseManager.DEFAULT_PHASE_IF_FAILURE_DESCRIPTION),
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       22,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       22,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       22,
                                       phase_second,
                                       PhaseManager.DEFAULT_PHASE_DESCRIPTION),
                       ]

        timeline = Timeline()
        timeline.record_list = record_list
        simulation_results = SimulationResults(5, 25)
        simulation_results.update_with_des_results({'number_of_time_iteration':5}, timeline)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].min, 5)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].max, 5)
        self.assertEqual(simulation_results.execution_metrics_statistics['number_of_time_iteration'].sum, 5)

        result_record_phase_first_init = ResultRecordEntryPhase(phase_first,RecordGenerator.DESCRIPTION_INIT)
        result_record_phase_first = ResultRecordEntryPhase(phase_first,PhaseManager.DEFAULT_PHASE_DESCRIPTION)
        result_record_phase_second = ResultRecordEntryPhase(phase_second,PhaseManager.DEFAULT_PHASE_DESCRIPTION)
        result_record_basic1_running_phase_first_init = ResultRecordEntryComponent(basic_1, Status.RUNNING, phase_first,RecordGenerator.DESCRIPTION_INIT)
        result_record_basic1_running_phase_first = ResultRecordEntryComponent(basic_1, Status.RUNNING, phase_first,"Repair")
        result_record_basic1_failed_phase_first = ResultRecordEntryComponent(basic_1, Status.FAILED, phase_first,"Failure")
        result_record_basic1_under_repair_phase_first = ResultRecordEntryComponent(basic_1, Status.UNDER_REPAIR, phase_first,"Repair")
        result_record_basic2 = ResultRecordEntryComponent(basic_2, Status.RUNNING, phase_first,RecordGenerator.DESCRIPTION_INIT)
        result_record_compound = ResultRecordEntryComponent(compound, Status.RUNNING, phase_first,RecordGenerator.DESCRIPTION_INIT)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_phase_first_init].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_phase_first].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_phase_second].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic1_running_phase_first_init].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic1_running_phase_first].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic1_failed_phase_first].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic1_under_repair_phase_first].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_basic2].min, 1)
        self.assertEqual(simulation_results.occurrence_statistics[result_record_compound].min, 1)

        self.assertEqual(simulation_results.duration_statistics[result_record_phase_first_init].min, 10)
        self.assertEqual(simulation_results.duration_statistics[result_record_phase_first].min, 2)
        self.assertEqual(simulation_results.duration_statistics[result_record_basic1_running_phase_first_init].min, 10)
        self.assertEqual(simulation_results.duration_statistics[result_record_basic1_failed_phase_first].min, 0)  # immediate repair
        self.assertEqual(simulation_results.duration_statistics[result_record_basic1_under_repair_phase_first].min, 1)
        self.assertEqual(simulation_results.duration_statistics[result_record_basic2].min, 10)
        self.assertEqual(simulation_results.duration_statistics[result_record_compound].min, 10)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
import os
from _datetime import datetime

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.system_utils import SystemUtils
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.exporter.xlsx_result_exporter_component_listing import XLSXResultExporterListing
from availsim4core.src.exporter.xlsx_result_exporter_component_tree_extented import \
    XLSXResultExporterComponentTreeExtended
from availsim4core.src.exporter.xlsx_result_exporter_component_tree_simple import XLSXResultExporterComponentTreeSimple
from availsim4core.src.exporter.xlsx_result_exporter_connectivity import XLSXResultExporterConnectivityMatrix
from availsim4core.src.exporter.xlsx_result_exporter_execution_metrics import XLSXResultExporterExecutionMetrics
from availsim4core.src.exporter.xlsx_result_exporter_last_timeline import XLSXResultExporterLastTimeline
from availsim4core.src.exporter.xlsx_result_exporter_summary import XLSXResultExporterSummary
from availsim4core.src.exporter.xlsx_result_exporter_rca import XLSXResultExporterRCA
from availsim4core.src.exporter.xlsx_result_exporter_critical_failure_paths import XLSXResultsExporterCriticalFailurePaths
from availsim4core.src.results.results import Results


class DiagnosticType:
    DIAGNOSTIC_EXPORTER = {
        "EXECUTION_METRICS": XLSXResultExporterExecutionMetrics,
        "SUMMARY": XLSXResultExporterSummary,
        "LAST_TIMELINE": XLSXResultExporterLastTimeline,
        "RCA": XLSXResultExporterRCA,
        "COMPONENT_TREE_SIMPLE": XLSXResultExporterComponentTreeSimple,
        "COMPONENT_TREE_EXTENDED": XLSXResultExporterComponentTreeExtended,
        "CONNECTIVITY_MATRIX": XLSXResultExporterConnectivityMatrix,
        "COMPONENT_LISTING": XLSXResultExporterListing,
        "GRAPH" : None,
        "CRITICAL_FAILURE_PATHS": None,
    }


class AggregateExporter(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str = None):

        # create the output directory if needed
        os.makedirs(output_folder, exist_ok=True)

        if unique_output_file_identifier_str is None:
            unique_output_file_identifier_str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        for diagnostic in analysis.simulation.get_list_of_diagnosis():
            logging.debug(f"Exporting {diagnostic}")

            if diagnostic == "GRAPH":
                # avoiding to import that module if it's not need because that module is a bit exotic
                from availsim4core.src.exporter.xlsx_result_exporter_graph import XLSXResultExporterGraph
                exporter = XLSXResultExporterGraph

            elif "CRITICAL_FAILURE_PATHS" in diagnostic:
                if SystemUtils.is_string_containing_parenthesis(diagnostic):
                    root_component_for_the_analysis = SystemUtils.extract_arguments_within_parenthesis(diagnostic)
                    logging.info(f"The diagnostic CRITICAL_FAILURE_PATHS is performed using the root component "
                                 f"{root_component_for_the_analysis} as provided with the argument {diagnostic}")
                else:
                    root_component_for_the_analysis = context.root_component.name
                    logging.info(f"The diagnostic CRITICAL_FAILURE_PATHS is performed using the root component "
                                 f"{root_component_for_the_analysis} as a default value because no speficic value was "
                                 f"provided with the argument {diagnostic}. This can be performed by adding the name of "
                                 f"the component which should be used as root between parenthesis, such as: "
                                 f"CRITICAL_FAILURE_PATHS(specific_component_name)")

                exporter = XLSXResultsExporterCriticalFailurePaths(root_component_for_the_analysis)

            else:
                exporter = DiagnosticType.DIAGNOSTIC_EXPORTER[diagnostic]

            exporter.export(context,
                            output_folder,
                            analysis,
                            results,
                            execution_time,
                            unique_output_file_identifier_str)


        # Default exporter export last in order to be sure previous exporters had the time to finish properly
        DiagnosticType.DIAGNOSTIC_EXPORTER["EXECUTION_METRICS"].export(context,
                                                                       output_folder,
                                                                       analysis,
                                                                       results,
                                                                       execution_time,
                                                                       unique_output_file_identifier_str)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.system_utils import SystemUtils


class test_SystemUtils(unittest.TestCase):

    def test__is_shared_child_true(self):
        shared_child_name = "A(B)"
        self.assertTrue(SystemUtils.is_string_containing_parenthesis(shared_child_name))

    def test__is_shared_child_false(self):
        shared_child_name = "A"
        self.assertFalse(SystemUtils.is_string_containing_parenthesis(shared_child_name))

    def test__extract_common_parent_name(self):
        shared_child_name = "A(B)"
        self.assertEqual("B", SystemUtils.extract_arguments_within_parenthesis(shared_child_name))

    def test__extract_shared_child_name(self):
        shared_child_name = "A(B)"
        self.assertEqual("A", SystemUtils.extract_name_of_function_from_string(shared_child_name))

    def test_find_first_lowest_level_ancestor_with_name__single_parent(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_1 = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        component = Compound(3, "C", 2, [parent_1], ChildrenLogic(), [])
        lowest_scope_common_ancestor_name = "ROOT_NODE"

        result = SystemUtils.find_first_lowest_level_ancestor_with_name(component, lowest_scope_common_ancestor_name)

        self.assertEqual(root_node, result)

    def test_find_first_lowest_level_ancestor_with_name__own_ancestor(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        lowest_scope_common_ancestor_name = "ROOT_NODE"

        result = SystemUtils.find_first_lowest_level_ancestor_with_name(root_node, lowest_scope_common_ancestor_name)

        self.assertEqual(root_node, result)

    def test_find_first_lowest_level_ancestor_with_name__multiple_parents(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_1 = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        parent_3 = Compound(3, "B", 2, [root_node], ChildrenLogic(), [])
        component = Compound(4, "C", 3, [parent_1, parent_3], ChildrenLogic(), [])
        lowest_scope_common_ancestor_name = "ROOT_NODE"

        result = SystemUtils.find_first_lowest_level_ancestor_with_name(component, lowest_scope_common_ancestor_name)

        self.assertEqual(root_node, result)

    def test_find_first_lowest_level_ancestor_with_name__check_right_path_tree(self):
        """
                       ROOT_NODE
                       //     \\
         LEVEL 1       A       Z  --> expected
                       |       |
         LEVEL 2       B       D
                       \\     //
                         BASIC
        """
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_level_1_left = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        parent_level_1_right = Compound(3, "Z", 2, [root_node], ChildrenLogic(), [])
        parent_level_2_left = Compound(4, "B", 2, [parent_level_1_left], ChildrenLogic(), [])
        parent_level_2_right = Compound(5, "D", 2, [parent_level_1_right], ChildrenLogic(), [])
        basic = Compound(6, "BASIC", 2, [parent_level_2_left, parent_level_2_right], ChildrenLogic(), [])
        lowest_scope_common_ancestor_name = "Z"

        result = SystemUtils.find_first_lowest_level_ancestor_with_name(basic, lowest_scope_common_ancestor_name)

        self.assertEqual(parent_level_1_right, result)

    def test_find_first_lowest_level_ancestor_with_name__check_parents_same_name(self):
        """
                        ROOT_NODE
                        //    \\
         LEVEL 1       A       ROOT_NODE  --> expected result
                        \\   //
                         BASIC
        """
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_level_1_left = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        parent_level_1_right = Compound(3, "ROOT_NODE", 2, [root_node], ChildrenLogic(), [])
        basic = Compound(6, "BASIC", 2, [parent_level_1_left, parent_level_1_right], ChildrenLogic(), [])
        lowest_scope_common_ancestor_name = "ROOT_NODE"

        result = SystemUtils.find_first_lowest_level_ancestor_with_name(basic, lowest_scope_common_ancestor_name)

        self.assertEqual(parent_level_1_right, result)

    def test_find_first_lowest_level_ancestor_with_name__two_nodes_same_name_different_scopes(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])

        lowest_common_parent_1 = Compound(2, "B", 1, [root_node], ChildrenLogic(), [])
        parent_1 = Compound(3, "D", 1, [lowest_common_parent_1], ChildrenLogic(), [])
        parent_2 = Compound(4, "E", 1, [lowest_common_parent_1], ChildrenLogic(), [])
        shared_component_1 = Compound(5, "C", 1, [parent_1, parent_2], ChildrenLogic(), [])

        lowest_common_parent_2 = Compound(6, "B", 2, [root_node], ChildrenLogic(), [])
        parent_3 = Compound(7, "D", 1, [lowest_common_parent_2], ChildrenLogic(), [])
        parent_4 = Compound(8, "E", 1, [lowest_common_parent_2], ChildrenLogic(), [])
        shared_component_2 = Compound(9, "C", 1, [parent_3, parent_4], ChildrenLogic(), [])

        lowest_scope_common_ancestor_name = "B"

        result_shared_component_1 = SystemUtils.find_first_lowest_level_ancestor_with_name(shared_component_1,
                                                                                           lowest_scope_common_ancestor_name)

        self.assertEqual(lowest_common_parent_1, result_shared_component_1)

        result_shared_component_2 = SystemUtils.find_first_lowest_level_ancestor_with_name(shared_component_2,
                                                                                           lowest_scope_common_ancestor_name)

        self.assertEqual(lowest_common_parent_2, result_shared_component_2)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.reader.xlsx.xlsx_reader import XLSXReader
from availsim4core.src.sensitivity_analysis.exploration_strategy_factory import ExplorationStrategyFactory
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy_director import \
    SystemModifierCombinationStrategyDirector


class SensitivityAnalysisSheet:
    SHEET = "SENSITIVITY_ANALYSIS"
    EXPLORATION_STRATEGY = "EXPLORATION_STRATEGY"
    PARAMETER_NAME = "PARAMETER_NAME"
    VALUES = "VALUES"


class SensitivityAnalysisError(Exception):
    pass


class SensitivityAnalysisReader(XLSXReader):
    """
    Specific reader for the sensitivity analysis.
    """

    @classmethod
    def generate_system_modifier_combination_list(cls,
                                                  sensitivity_analysis_file_path):
        """
        Given a path for the configuration file of the sensitivity analysis it returns a list of
        SystemModifierCombination to apply.
        :param sensitivity_analysis_file_path the path of the configuration file for the sensitivity analysis.
        """
        exploration_strategy_list = cls._read_exploration_strategy_list(sensitivity_analysis_file_path)
        return SystemModifierCombinationStrategyDirector.execute(exploration_strategy_list)

    @classmethod
    def _read_exploration_strategy_list(cls,
                                        sensitivity_analysis_file_path):
        """
        Reads the sensitivity configuration file and returns the list of strategies to apply.
        """
        sensitivity_analysis_dictionary = cls.read(sensitivity_analysis_file_path)

        exploration_strategy_list = []
        for row in sensitivity_analysis_dictionary[SensitivityAnalysisSheet.SHEET].values():

            exception_message_hint = f"sensitivity analysis row: {row}"

            current_line_strategy = cls.clean_cell_str_else_raise(row[SensitivityAnalysisSheet.EXPLORATION_STRATEGY],
                                                  exception_message_hint=exception_message_hint)
            current_line_parameter_name = cls.clean_cell_str_else_raise(row[SensitivityAnalysisSheet.PARAMETER_NAME],
                                                        exception_message_hint=exception_message_hint)
            current_line_values = row[SensitivityAnalysisSheet.VALUES]
            current_line_values = [current_line_values] if isinstance(current_line_values,
                                                                      (int, float)) \
                else cls.extract_from_str_list(current_line_values)

            if len(current_line_values) == 0:
                message_exception = f"Error in the sensitivity analysis file: \n" \
                                    f"The parameter {current_line_parameter_name} is associated to an empty list of values"
                logging.exception(message_exception)
                raise SensitivityAnalysisError(message_exception)

            strategy = ExplorationStrategyFactory.build(current_line_parameter_name,
                                                        current_line_values,
                                                        current_line_strategy)
            exploration_strategy_list.append(strategy)

        logging.debug(f"Extracted from sensitivity file exploration_strategy_list: {exploration_strategy_list}")
        return exploration_strategy_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.component import Component

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event_factory import \
    FailureEventFactory
from availsim4core.src.discrete_event_simulation.event.b_event.inspection_event.inspection_event_factory import \
    InspectionEventFactory
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import NextPhaseEvent


class BEventGenerator:
    """
    This static class provides methods to initialise the BEvents of the simulation.
    """

    @classmethod
    def generate_first_events(cls,
                              absolute_simulation_time: float,
                              context: Context):
        """
        Function generating the first B events when a simulation is started.
        Those events are generated by the first failures of components, periodic inspections, first change of phases etc.
        """
        set_of_b_events = set()
        set_of_b_events.update(
            cls.get_b_events_inspection(
                absolute_simulation_time,
                context,
                context.root_component
            )
        )
        set_of_b_events.update(
            cls._generate_first_phase_events(
                absolute_simulation_time, context
            )
        )
        return set_of_b_events

    @staticmethod
    def _generate_first_phase_events(absolute_simulation_time: float,
                                     context: Context):
        """
        Function generating the first B events related to phase changes
        """
        next_phase_time = absolute_simulation_time + context.phase_manager.current_phase.law.get_random_value()
        event = NextPhaseEvent(next_phase_time,
                               context)
        return {event}

    @classmethod
    def get_b_events_inspection(cls,
                                absolute_simulation_time: float,
                                context: Context,
                                component: Component):
        """
        Function generating first inspection B events of a component recursively,
        calling itself on children of the current component if any
        """
        set_of_events = set()
        for child_component in component.get_children():
            events = cls.get_b_events_inspection(absolute_simulation_time,
                                                 context,
                                                 child_component)
            set_of_events.update(events)
        if not component.get_children():
            inspection_event = InspectionEventFactory.build(absolute_simulation_time,
                                                            context,
                                                            component,
                                                            component.failure_mode)
            if inspection_event is not None:
                set_of_events.add(inspection_event)

        return set_of_events



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_if_failure_event import \
    NextPhaseIfFailureEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent


class OrderNextPhaseIfFailureEvent(CEvent):
    """
    CEvent class in charge of creating a b_event {NextPhaseIfFailureEvent}.
    Note: The valid condition is based on the ROOT component of the system.
    # TODO removed the condition on the ROOT component. the condition should be recursive on any component of the system.
    """
    __slots__ = 'failure_mode'

    def __init__(self,
                 priority: int,
                 context: Context,
                 failure_mode: FailureMode):
        super().__init__(priority, context)
        self.failure_mode = failure_mode

    def __eq__(self, other):
        if not isinstance(self, OrderNextPhaseIfFailureEvent):
            return NotImplemented
        return super().__eq__(other) and self.failure_mode == other.failure_mode

    def __hash__(self):
        return hash((type(self), self.priority, self.failure_mode))

    def __str__(self):
        return f"OrderNextPhaseIfFailureEvent:: " \
               f"priority:{self.priority} - " \
               f"failure_mode:{self.failure_mode.name}"

    def is_condition_valid(self):
        if self.failure_mode.phase_change_trigger in ["AFTER_REPAIR"]:
            return self.context.root_component.status == Status.RUNNING
        elif self.failure_mode.phase_change_trigger in ["AFTER_FAILURE"]:
            return True
        elif self.failure_mode.phase_change_trigger in ["NEVER"]:
            return False

    def generate_b_events(self, absolute_simulation_time):
        next_phase_time = absolute_simulation_time
        event = NextPhaseIfFailureEvent(next_phase_time,
                                        self.context,
                                        self.failure_mode)
        return {event}



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.sensitivity_analysis.exploration_strategy import Zip
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy import \
    SystemModifierCombinationStrategy


class ZipStrategyError(Exception):
    pass


class SystemModifierCombinationZipStrategy(SystemModifierCombinationStrategy):
    """
    Strategy to apply the Inner logic.

    :eg.
        Default analysis with
            - A:0, B:0, C:0
        Given:
            - Zip(A; 1,2,3)
            - Zip(B; 1,2,3)
            - Zip(C; 1,2,3)

        It results in 3 analysis :
            - A:1, B:1, C:1
            - A:2, B:2, C:2
            - A:3, B:3, C:3
    """

    @classmethod
    def execute(cls, exploration_strategy_list):
        """
        Given a list of {ExplorationStrategy} it returns a list of corresponding {SystemModifierCombination}.
        :param exploration_strategy_list the list from which the Zip strategy will be extracted and applied
        to get the corresponding combination
        :returns list of {SystemModifierCombination} corresponding to the Inner logic.
        """
        zip_exploration_strategy_list = cls.filter_exploration(exploration_strategy_list, Zip)
        zip_combination_list = []

        # if len(zip_exploration_strategy_list)==0, then the user did not request any ZIP strategy

        if len(zip_exploration_strategy_list)==1:
            exception_message = (f"ZIP strategies misconfigured, at least two occurrences of ZIP strategies should be "
                                 f"in the sensitivity analysis input file but only one is present")
            logging.exception(exception_message)
            raise ZipStrategyError(exception_message)

        elif len(zip_exploration_strategy_list)>1:

            # testing if the length of all the zip_exploration_strategy are equal
            length = len(zip_exploration_strategy_list[0].values)
            for idx in range(1,len(zip_exploration_strategy_list)):
                idx_length = len(zip_exploration_strategy_list[idx].values)
                if idx_length != length:
                    exception_message = (f"ZIP strategies missconfigured, the length of the {idx} strategy has "
                                         f"{idx_length} elements instead of {length}")
                    logging.exception(exception_message)
                    raise ZipStrategyError(exception_message)

            # building the system modifiers
            for idv in range(0,length):
                curr_system_modifier_list = [SystemModifier(zip_exploration_strategy_list[0].parameter_name,
                                                            zip_exploration_strategy_list[0].values[idv])]
                for idx in range(1,len(zip_exploration_strategy_list)):
                    curr_system_modifier_list.append(SystemModifier(zip_exploration_strategy_list[idx].parameter_name,
                                                                    zip_exploration_strategy_list[idx].values[idv]))
                zip_combination_list.append(SystemModifierCombination(curr_system_modifier_list))

        return zip_combination_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnitFactory, \
    MinimalReplaceableUnit
from availsim4core.src.context.system.probability_law.probability_law_factory import ProbabilityLawFactory


class test_MinimalReplaceableUnit(unittest.TestCase):

    def test_build_with_one_element(self):
        # GIVEN
        minimal_replaceable_unit_name = "fakeMRU"
        minimal_replaceable_unit_repair_law_name = "FIX"
        minimal_replaceable_unit_repair_parameters = [0]
        minimal_replaceable_repair_schedule = "IMMEDIATE"
        triggering_status_list_of_str = ["FAILED"]
        lowest_common_parent_name_str = "fakeParent"

        # WHEN
        mru_list_generated = MinimalReplaceableUnitFactory.build(minimal_replaceable_unit_name,
                                                                 minimal_replaceable_unit_repair_law_name,
                                                                 minimal_replaceable_unit_repair_parameters,
                                                                 minimal_replaceable_repair_schedule,
                                                                 triggering_status_list_of_str,
                                                                 lowest_common_parent_name_str)

        # THEN
        mru_list_expected = [MinimalReplaceableUnit(minimal_replaceable_unit_name,
                                                    ProbabilityLawFactory.build(
                                                        minimal_replaceable_unit_repair_law_name,
                                                        minimal_replaceable_unit_repair_parameters),
                                                    minimal_replaceable_repair_schedule,
                                                    Status.FAILED,
                                                    lowest_common_parent_name_str)]

        self.assertListEqual(mru_list_generated, mru_list_expected)

    def test_build_with_list(self):
        # GIVEN
        minimal_replaceable_unit_name = "fakeMRU"
        minimal_replaceable_unit_repair_law_name = "FIX"
        minimal_replaceable_unit_repair_parameters = [0]
        minimal_replaceable_repair_schedule = "IMMEDIATE"
        triggering_status_list_of_str = ["FAILED", "BLIND_FAILED"]
        lowest_common_parent_name_str = "fakeParent"

        # WHEN
        mru_list_generated = MinimalReplaceableUnitFactory.build(minimal_replaceable_unit_name,
                                                                 minimal_replaceable_unit_repair_law_name,
                                                                 minimal_replaceable_unit_repair_parameters,
                                                                 minimal_replaceable_repair_schedule,
                                                                 triggering_status_list_of_str,
                                                                 lowest_common_parent_name_str)

        # THEN
        mru_list_expected = [MinimalReplaceableUnit(minimal_replaceable_unit_name,
                                                    ProbabilityLawFactory.build(
                                                        minimal_replaceable_unit_repair_law_name,
                                                        minimal_replaceable_unit_repair_parameters),
                                                    minimal_replaceable_repair_schedule,
                                                    Status.FAILED,
                                                    lowest_common_parent_name_str),
                             MinimalReplaceableUnit(minimal_replaceable_unit_name,
                                                    ProbabilityLawFactory.build(
                                                        minimal_replaceable_unit_repair_law_name,
                                                        minimal_replaceable_unit_repair_parameters),
                                                    minimal_replaceable_repair_schedule,
                                                    Status.BLIND_FAILED,
                                                    lowest_common_parent_name_str)
                             ]

        self.assertListEqual(mru_list_generated, mru_list_expected)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import copy
import logging
from datetime import datetime

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.discrete_event_simulation import DiscreteEventSimulation
from availsim4core.src.results.simulation_results import SimulationResults
from availsim4core.src.simulation.monte_carlo import MonteCarlo
from availsim4core.src.simulation.des_random_generator.quasi_monte_carlo_generator import QuasiMonteCarloGenerator

from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.simulation.simulation import SimulationType


class QuasiMonteCarlo(MonteCarlo):
    """
    Defines a MonteCarlo simulation that will be triggered by the Sensitivity Analysis.
    Attributes:
    - minimum_number_of_simulations: minimun number of call to discrete event simulation.
    - maximum_number_of_simulations: maximun number of call to discrete event simulation.
    - convergence_margin: Evaluate on the fly the appropriate number of simulation to run.
    - maximum_execution_time: used on the Cluster to stop the simulation after this period of time.
    - seed: Control the random generator to get different values over each simulation. Deterministic approach.
    - list_of_diagnosis: Defines the metrics to compute.
    - result_simulation: The ResultSimulation to return.
    Extends the simulation, by implementing the run() method.
    """
    SIMULATION_TYPE = SimulationType.QUASI_MONTE_CARLO


    def run(self, context: Context):
        simulation_results = SimulationResults(self.maximum_number_of_simulations,
                                               self.duration_of_simulation)

        basic_components = [component for component in context.root_component.to_set() if isinstance(component, Basic)]
        qp_sobol = QuasiMonteCarloGenerator(basic_components, self.duration_of_simulation)
        qp_sobol_samples = qp_sobol.generate_samples(self.seed, self.seed+self.maximum_number_of_simulations)
        simulation_weight = 1/self.maximum_number_of_simulations

        for simulation_index in range(0, self.maximum_number_of_simulations):
            start_time = datetime.now()
            logging.info(f"Starting Quasi-Monte Carlo simulation {simulation_index+1} / {self.maximum_number_of_simulations}")

            random_sequence = qp_sobol_samples[simulation_index]
            qp_sobol.set_ttfs_of_failure_modes(random_sequence)

            local_context = copy.deepcopy(context)

            # initialisation of a DES
            simulation_des = DiscreteEventSimulation(self.seed + simulation_index,
                                                     self.duration_of_simulation,
                                                     local_context, None, -1 , 0)

            # simulation
            _, _ = simulation_des.run()
            timeline_record = local_context.timeline_record

            # updating statistics
            self.update_statistics(simulation_results, 
                                   simulation_des, 
                                   start_time, 
                                   simulation_weight,
                                   0)

        simulation_results.evaluate_result()
        simulation_results.last_simulation_timeline = timeline_record.record_list
        return simulation_results



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest
import pandas
import PIL
import numpy as np

from availsim4core import main

class test_fancy_exporter(unittest.TestCase):
    param_scenario = [
        ["./availsim4core/test/E2E/input/fancy_exporter/simulation.xlsx",
         "./availsim4core/test/E2E/input/fancy_exporter/system.xlsx",
         "./availsim4core/test/E2E/input/fancy_exporter/theoretical_output.xlsx"],
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, theoretical_result_file in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 theoretical_result_file)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    theoretical_result_file):

        output_folder = "./availsim4core/test/E2E/output/fancy_exporter/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, None, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")

        for sheet_name in ['RESULTS_COMPONENT_TREE_EXTENDED','RESULTS_COMPONENT_TREE_SIMPLE',
                           'RESULTS_CONNECTIVITY_MATRIX', 'RESULTS_COMPONENT_LISTING',
                           'RESULTS_CRITICAL_FAILURE_PATHS']:
            actual_output = pandas.read_excel(result_simulation_file_list[0], sheet_name=sheet_name, engine='openpyxl')
            theoretical_output = pandas.read_excel(theoretical_result_file, sheet_name=sheet_name, engine='openpyxl')

            self.assertTrue(actual_output.equals(theoretical_output))

        # compare the RESULTS_GRAPH
        ## TODO:
        ## equality TEST not passing the CI, the picture have slightly different layout, I don't know why...
        #
        #theoretical_png = np.array(PIL.Image.open("./availsim4core/test/E2E/input/fancy_exporter/theoretical_tree.png"))
        #actual_png = np.array(PIL.Image.open(result_simulation_file_list[0].split('.xlsx')[0]+'.png'))
        #self.assertTrue(np.array_equal(theoretical_png, actual_png))
        #
        #theoretical_linux_png = np.array(PIL.Image.open("./availsim4core/test/E2E/input/fancy_exporter/theoretical_tree_linux.png"))
        #actual_linux_png = np.array(PIL.Image.open(result_simulation_file_list[0].split('.xlsx')[0]+'_linux.png'))
        #self.assertTrue(np.array_equal(theoretical_linux_png, actual_linux_png))


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
import os

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results

import networkx as nx
import matplotlib.pyplot as plt


class XLSXResultExporterGraph(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        def graph_export(root: Component):

            def get_name(component: Component):
                """
                :param component:
                :return: string identifying a component by its name
                """
                name = f"{component.name}_{component.local_id}_{component.global_id}"
                if component.list_of_mru_trigger:
                    name+= '\nTRIGGER MRU: '+"".join([x.name+', ' for x in component.list_of_mru_trigger])
                if not component.get_children() and component.list_of_mru_group:
                    name+= '\nIN MRU: '+"".join([x.name+', ' for x in component.list_of_mru_group])
                return name

            def get_edge(parent: Component,
                         list_of_edges,
                         shell_dict,
                         level=0):
                """
                Recursive function defining the list of edges and the depth of each node
                :param parent: component studied
                :param list_of_edges: accumulated list of edges
                :param shell_dict: accumulated dictionaries of the different node (component) in the different layers (shell)
                :param level: depth of the component (use for the shell)
                :return:
                """

                parent_name = get_name(parent)

                if level in shell_dict.keys():
                    # extend the current level
                    shell_dict[level].append(parent_name)
                else:
                    # instantiate the level with a list containing one element
                    shell_dict[level] = [parent_name]

                for child in parent.get_children():
                    child_name = get_name(child)
                    list_of_edges.append((parent_name, child_name))
                    list_of_edges, shell_dict = get_edge(child,
                                                         list_of_edges,
                                                         shell_dict,
                                                         level+1)

                return list_of_edges, shell_dict

            def generate_picture_with_linux_command(picture_format: str):
                """
                Function using the command "dot" present on linux os to generate a picture
                :param picture_format:
                :return:
                """
                output_dot = cls.OUTPUT_FILE_TEMPLATE.format(output_folder,
                                                             analysis.id,
                                                             unique_output_file_identifier_str
                                                             ).split(".xlsx")[0] + '_linux.' + picture_format
                command = f"dot -T{picture_format} {dot_filename} > {output_dot}"
                try:
                    os.system(command)
                except:
                    logging.warning(f"Impossible to generate {output_dot}")

            list_of_edges, shell_dict = get_edge(root,
                                                 [],
                                                 {0: []},
                                                 level=0)

            G = nx.DiGraph()
            G.add_edges_from(list_of_edges)
            # generating a dot file
            dot_filename = 'tmp_tree_file.dot'
            nx.drawing.nx_agraph.write_dot(G, dot_filename)
            # using the dot file to create picture directly from the linux os
            generate_picture_with_linux_command('png')
            generate_picture_with_linux_command('svg')

            number_of_layers_in_the_graph = len(shell_dict.keys())
            number_of_columns_in_the_graph = max([len(x) for x in shell_dict.values()])
            plt.figure(figsize=(3*number_of_columns_in_the_graph,
                                1*number_of_layers_in_the_graph))

            # generate a picture within python
            pos = nx.drawing.nx_agraph.graphviz_layout(G, prog='dot')
            labels = nx.draw_networkx_labels(
                G,
                pos,
                font_size=10,
                horizontalalignment="right",
                verticalalignment="top"
            )
            # rotate the labels by 15 degrees in order to improve the readability / limit the overlap between text
            for _, t in labels.items():
                t.set_rotation(15)

            nx.draw(
                G,
                pos=pos,
                labels=labels,
                with_labels=False,
                arrows=True,
                arrowsize=20,
                node_color='red',
                node_size=10,
                alpha=.25
            )

            output_python_png = cls.OUTPUT_FILE_TEMPLATE.format(
                output_folder,
                analysis.id,
                unique_output_file_identifier_str).split("xlsx")[0] + 'png'
            plt.savefig(output_python_png,dpi=300)

            output_python_svg = cls.OUTPUT_FILE_TEMPLATE.format(
                output_folder,
                analysis.id,
                unique_output_file_identifier_str).split("xlsx")[0] + 'svg'
            plt.savefig(output_python_svg)

        graph_export(context.root_component)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_convergence_rare_event(unittest.TestCase):
    """
    class used to recover analytical results: the probability of failure of simple systems after a given time,
    following an exponential distribution law of failure.
    For the system with 2oo4 logic and a mttf of 15 and a duration of simulation of 1, the result is 0.0010211...
    For the system with 3oo4 logic and a mttf of 5 and a duration of simulation of 1, the result is 0.0010797...
    """


    scenario_2oo4_MC = [
        "./availsim4core/test/E2E/input/convergence_rare_event/10_000_MC_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/2oo4_system.xlsx",
        None,
        1,
        [
            Result(0.0004, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    scenario_2oo4_QMC = [
        "./availsim4core/test/E2E/input/convergence_rare_event/10_000_QMC_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/2oo4_system.xlsx",
        None,
        1,
        [
            Result(0.0009, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    scenario_2oo4_restart = [
        "./availsim4core/test/E2E/input/convergence_rare_event/2oo4_restart_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/2oo4_system.xlsx",
        None,
        1,
        [
            Result(0.0008, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    scenario_3oo4_MC = [
        "./availsim4core/test/E2E/input/convergence_rare_event/10_000_MC_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/3oo4_system.xlsx",
        None,
        1,
        [
            Result(0.0015, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    scenario_3oo4_QMC = [
        "./availsim4core/test/E2E/input/convergence_rare_event/10_000_QMC_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/3oo4_system.xlsx",
        None,
        1,
        [
            Result(0.0009, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    scenario_3oo4_restart = [
        "./availsim4core/test/E2E/input/convergence_rare_event/3oo4_restart_simulation.xlsx",
        "./availsim4core/test/E2E/input/convergence_rare_event/3oo4_system.xlsx",
        None,
        1,
        [
            Result(0.00107, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
        ]
    ]

    param_scenario = [
        scenario_2oo4_MC,
        scenario_2oo4_QMC,
        scenario_2oo4_restart,
        scenario_3oo4_MC,
        scenario_3oo4_QMC,
        scenario_3oo4_restart,
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        # random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_mean_result in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file, system_file, sensitivity_file, expected_number_file_result,
                                 expected_mean_result)

    def _runner_E2E(self, simulation_file, system_file, sensitivity_file, expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/convergence_rare_event/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTrigger
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent


class OrderJumpPhaseEvent(CEvent):
    """
    CEvent class in charge of creating a b_event {JumpPhaseEvent} if the jump conditions are satisfied.
    """
    def __init__(self,
                 priority: int,
                 context: Context,
                 phase_jump_trigger: PhaseJumpTrigger):
        super().__init__(priority, context)
        self.phase_jump_trigger = phase_jump_trigger

    def __eq__(self, other):
        if not isinstance(self, OrderJumpPhaseEvent):
            return NotImplemented
        return super().__eq__(other) and \
            self.phase_jump_trigger == other.phase_jump_trigger

    def __hash__(self):
        return hash((type(self), self.priority, self.phase_jump_trigger))

    def __str__(self):
        return f"OrderJumpPhaseEvent:: " \
               f"priority:{self.priority}"

    def is_condition_valid(self):
        """
        The condition for the event to be valid in order to be processed.
        By default the value returned is True
        """
        for component in self.context.root_component.to_set():
            if component.name == self.phase_jump_trigger.component_name:
                if component.status == self.phase_jump_trigger.component_status:
                    if self.context.phase_manager.current_phase == self.phase_jump_trigger.from_phase:
                        return True
        return False

    def generate_b_events(self, absolute_simulation_time):
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.jump_phase_event import \
            JumpPhaseEvent
        jump_phase_time = absolute_simulation_time
        event = JumpPhaseEvent(jump_phase_time,
                               self.context,
                               self.phase_jump_trigger)
        return {event}



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.context.system.probability_law.exponential_law import ExponentialLaw
from availsim4core.src.context.system.probability_law.normal_law import NormalLaw
from availsim4core.src.context.system.probability_law.probability_law_factory import ProbabilityLawFactory, \
    ProbabilityLawFactoryError


class test_probabilityLawFactory(unittest.TestCase):

    def test_build_ExponentialLaw(self):
        distribution_str = "EXP"
        parameters = []
        result = ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertEqual(ExponentialLaw([]), result)

        distribution_str = "EXPONENTIAL"
        parameters = [2, 0]
        result = ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertEqual(ExponentialLaw([2, 0]), result)

    def test_build_NormalLaw(self):
        distribution_str = "NORMAL"
        parameters = []
        result = ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertEqual(NormalLaw([]), result)

    def test_build_DeterministicLaw(self):
        distribution_str = "FIX"
        parameters = []
        result = ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertEqual(DeterministicLaw([]), result)

        distribution_str = "DETERMINISTIC"
        parameters = [2, 3]
        result = ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertEqual(DeterministicLaw([2, 3]), result)

    def test_build_invalid_probability_law(self):
        distribution_str = "invalid"
        parameters = []
        with self.assertRaises(ProbabilityLawFactoryError) as context:
            ProbabilityLawFactory.build(distribution_str, parameters)
        self.assertTrue('wrong type of distribution function' in str(context.exception))



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.children_logic_factory import ChildrenLogicFactory, \
    ChildrenLogicFactoryError
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.context.system.children_logic.required_component import RequiredComponent
from availsim4core.src.context.system.children_logic.tolerated_fault import ToleratedFault


class test_ChildrenLogicFactory(unittest.TestCase):
    scenario_1 = ["AND", And]
    scenario_2 = ["10RC", RequiredComponent]
    scenario_3 = ["1OO2", Oo]
    scenario_4 = ["10TF", ToleratedFault]
    build_scenario = [scenario_1, scenario_2, scenario_3, scenario_4]

    def test_build(self):
        for children_logic_str, expected_children_logic_instance in self.build_scenario:
            with self.subTest():
                self._build(children_logic_str, expected_children_logic_instance)

    def _build(self, children_logic_str, expected_children_logic_instance):
        result = ChildrenLogicFactory.build(children_logic_str,None)
        self.assertIsInstance(result, expected_children_logic_instance)

    def test_build_exception(self):
        self.assertRaises(ChildrenLogicFactoryError, ChildrenLogicFactory.build, "invalid_children_logic_str", None)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterComponentTreeSimple(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        def component_tree_simple_export(component: Component, level=0):
            """
            Function recursively exploring the system in order to build a graphical representation of the system.
            This "simple" version of the tree exporter provides a "light" output. A "extended" version exists too.
            The function is called on the root component. Then if a component has children, it is called again for its
            children.
            """

            # line printed in the output for the current component
            ret = (f";" * (level) +
                   f"{component.name}_{component.local_id}_{component.global_id}\n")

            # if that component has children (then it is a Compound type), apply the same function to the children
            if component.get_children():
                for child in component._children:
                    ret += component_tree_simple_export(child,level + 1)

            return ret

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        tree_string = component_tree_simple_export(context.root_component)

        tree_dataframe = pandas.DataFrame([x.split(';') for x in tree_string.split('\n')])

        tree_dataframe.to_excel(writer, sheet_name='RESULTS_COMPONENT_TREE_SIMPLE', index=False)

        writer.save()




# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.reader.xlsx.simulation_reader import SimulationReader, SimulationNotFoundError, \
    DiagnosticNotFoundError
from availsim4core.src.simulation.monte_carlo import MonteCarlo


class test_SimulationReader(unittest.TestCase):

    def test_generate_simulation(self):
        simulation_dict = {'SIMULATION':
            {
                0: {'SIMULATION_TYPE': 'MONTE_CARLO',
                    'MIN_NUMBER_OF_SIMULATION': 100,
                    'MAX_NUMBER_OF_SIMULATION': 100,
                    'CONVERGENCE_MARGIN': 2,
                    'MAX_EXECUTION_TIME': 600,
                    'SEED': 1,
                    'DIAGNOSTICS': '[summary]',
                    'SIMULATION_DURATION': 1000}
            }
        }

        expected_result = MonteCarlo(100, 100, 2, 600, 1, ["SUMMARY"], 1000)

        with patch.object(SimulationReader, "read", return_value=simulation_dict):
            result = SimulationReader.generate_simulation("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_simulation_with_list_of_diagnostics(self):
        simulation_dict = {'SIMULATION':
            {
                0: {'SIMULATION_TYPE': 'MONTE_CARLO',
                    'MIN_NUMBER_OF_SIMULATION': 100,
                    'MAX_NUMBER_OF_SIMULATION': 100,
                    'CONVERGENCE_MARGIN': 2,
                    'MAX_EXECUTION_TIME': 600,
                    'SEED': 1,
                    'DIAGNOSTICS': '[summary,last_timeline]',
                    'SIMULATION_DURATION': 1000}
            }
        }

        expected_result = MonteCarlo(100, 100, 2, 600, 1, ["SUMMARY","LAST_TIMELINE"], 1000)

        with patch.object(SimulationReader, "read", return_value=simulation_dict):
            result = SimulationReader.generate_simulation("fake_path")

        self.assertEqual(expected_result, result)

    def test_generate_simulation__exception(self):
        simulation_dict = {'SIMULATION':
            {
                0: {'SIMULATION_TYPE': 'FAKE'}
            }
        }

        with patch.object(SimulationReader, "read", return_value=simulation_dict):
            self.assertRaises(SimulationNotFoundError,
                              SimulationReader.generate_simulation,
                              "") # TODO what to put here?

    def test_generate_simulation_with_none_existing_diagnostics__exception(self):
        simulation_dict = {'SIMULATION':
            {
                0: {'SIMULATION_TYPE': 'MONTE_CARLO',
                    'MIN_NUMBER_OF_SIMULATION': 100,
                    'MAX_NUMBER_OF_SIMULATION': 100,
                    'CONVERGENCE_MARGIN': 2,
                    'MAX_EXECUTION_TIME': 600,
                    'SEED': 1,
                    'DIAGNOSTICS': '[summary,42]',
                    'SIMULATION_DURATION': 1000}
            }
        }

        with patch.object(SimulationReader, "read", return_value=simulation_dict):
            self.assertRaises(DiagnosticNotFoundError,
                              SimulationReader.generate_simulation,
                              "") # TODO what to put here?



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.rca.rca_manager import RootCauseAnalysisManager
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.discrete_event_simulation.discrete_event_simulation import DiscreteEventSimulation
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.blind_failure_event import \
    BlindFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.inspection_event.start_inspection_event import \
    StartInspectionEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.end_repairing_event import EndRepairingEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.start_repairing_event import \
    StartRepairingEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.inspection_event.order_end_inspection_event import \
    OrderEndInspectionEvent
from availsim4core.src.discrete_event_simulation.event.c_event.repair_event.order_repair_event import OrderRepairEvent
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.record_generator import RecordGenerator
from availsim4core.src.timeline.record_phase import RecordPhase


class test_DiscreteEventSimulation(unittest.TestCase):

    def test_run(self):
        phase = PhaseManager.NO_PHASE

        failure_mode_1 = FailureMode("test_run",
                                     DeterministicLaw([1]),
                                     DeterministicLaw([1]),
                                     Failure(FailureType.DETECTABLE),
                                     [phase],
                                     None,
                                     [phase],
                                     None,
                                     'AFTER_REPAIR',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        _basic_name_and_ids = f"{_basic.name}_{_basic.local_id}_{_basic.global_id}"
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager, set(), set())

        discrete_event_simulation = DiscreteEventSimulation(0, 3, context)

        _,_ = discrete_event_simulation.run()

        expected_root_node_timeline = RecordGenerator.get_initial_records(0, context)
        expected_root_node_timeline.extend([
            RecordComponent(_basic, Status.FAILED, 1, phase, f"test_run failure mode of component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.FAILED, 1, phase, f"test_run failure mode of component {_basic_name_and_ids}"),
            RecordComponent(_basic, Status.UNDER_REPAIR, 1, phase, f"UNDER REPAIR - component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.UNDER_REPAIR, 1, phase, f"UNDER REPAIR - component {_basic_name_and_ids}"),
            RecordComponent(_basic, Status.RUNNING, 2, phase, f"REPAIRED - component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.RUNNING, 2, phase, f"REPAIRED - component {_basic_name_and_ids}"),
            RecordPhase(phase, 2, PhaseManager.DEFAULT_PHASE_IF_FAILURE_DESCRIPTION),
            RecordComponent(root_component, Status.RUNNING, 2, phase, "Update following a phase change"),
            RecordComponent(_basic, Status.RUNNING, 2, phase, "Update following a phase change"),
            RecordComponent(_basic, Status.FAILED, 3, phase, f"test_run failure mode of component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.FAILED, 3, phase, f"test_run failure mode of component {_basic_name_and_ids}"),
            RecordComponent(_basic, Status.UNDER_REPAIR, 3, phase, f"UNDER REPAIR - component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.UNDER_REPAIR, 3, phase, f"UNDER REPAIR - component {_basic_name_and_ids}")
        ])

        self.assertEqual(root_component.status, Status.UNDER_REPAIR)
        self.assertEqual(_basic.status, Status.UNDER_REPAIR)
        self.assertEqual(discrete_event_simulation.context.timeline_record.record_list,
                         expected_root_node_timeline)

    def test_run__blind_failures(self):
        phase = Phase("phase_test", DeterministicLaw([10]), True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test_run",
                                     DeterministicLaw([2]),
                                     DeterministicLaw([2]),
                                     Failure(FailureType.BLIND),
                                     [phase],
                                     None,
                                     [phase],
                                     None,
                                     'NEVER',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        _basic_name_and_ids = f"{_basic.name}_{_basic.local_id}_{_basic.global_id}"
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager, set(), set())

        discrete_event_simulation = DiscreteEventSimulation(0, 5, context)

        _, _ = discrete_event_simulation.run()

        expected_root_node_timeline = RecordGenerator.get_initial_records(0, context)
        expected_root_node_timeline.extend([
            RecordComponent(_basic, Status.BLIND_FAILED, 2, phase, f"test_run failure mode of component {_basic_name_and_ids}"),
            RecordComponent(root_component, Status.BLIND_FAILED, 2, phase, f"test_run failure mode of component {_basic_name_and_ids}")])
        self.assertEqual(root_component.status, Status.BLIND_FAILED)
        self.assertEqual(_basic.status, Status.BLIND_FAILED)
        self.assertListEqual(discrete_event_simulation.context.timeline_record.record_list, expected_root_node_timeline)

    def test_run__inspection_events(self):
        inspection_name = "long_shut_down"
        inspection_period = 5
        inspection_duration = 1

        phase = Phase("phase_test", DeterministicLaw([10]), True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test_run",
                                     DeterministicLaw([2]),
                                     DeterministicLaw([2]),
                                     Failure(FailureType.BLIND),
                                     [phase],
                                     Inspection(inspection_name, inspection_period, inspection_duration),
                                     [phase],
                                     None,
                                     'NEVER',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        _basic_name = f"{_basic.name}_{_basic.local_id}_{_basic.global_id}"
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager, set(), set())
        discrete_event_simulation = DiscreteEventSimulation(0,
                                                            inspection_period + inspection_duration,
                                                            context)

        _,_ = discrete_event_simulation.run()

        expected_root_node_timeline = RecordGenerator.get_initial_records(0, context)
        expected_root_node_timeline.extend([
            RecordComponent(_basic, Status.BLIND_FAILED, 2, phase, f"test_run failure mode of component {_basic_name}"),
            RecordComponent(root_component, Status.BLIND_FAILED, 2, phase, f"test_run failure mode of component {_basic_name}"),
            RecordComponent(_basic, Status.INSPECTION, 5, phase, f"INSPECTION"),
            RecordComponent(root_component, Status.INSPECTION, 5, phase, f"INSPECTION"),
            RecordComponent(_basic, Status.RUNNING, 6, phase, f"EndInspectionEvent"),
            RecordComponent(root_component, Status.RUNNING, 6, phase, f"EndInspectionEvent")])
        self.assertEqual(root_component.status, Status.RUNNING)
        self.assertEqual(_basic.status, Status.RUNNING)
        self.assertListEqual(discrete_event_simulation.context.timeline_record.record_list, expected_root_node_timeline)

    def test__execute_b_event__detectable_failure(self):
        phase = Phase("phase_test", DeterministicLaw([10]),  True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test__execute_b_event",
                                     DeterministicLaw([1]),
                                     DeterministicLaw([1]),
                                     Failure(FailureType.DETECTABLE),
                                     [phase],
                                     None,
                                     [phase],
                                     None,
                                     'AFTER_REPAIR',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager, set(), set())

        detectable_failure_event = DetectableFailureEvent(0, context, _basic, failure_mode_1)

        discrete_event_simulation = DiscreteEventSimulation(0, 3, context)

        discrete_event_simulation._execute_b_event(detectable_failure_event)

        self.assertEqual(_basic.status, Status.FAILED)

        expected_c_events_set = set([OrderRepairEvent(
            priority=CEventPriority.ORDER_REPAIR_EVENT,
            context=context,
            component=_basic,
            event=detectable_failure_event,
            failure_mode=failure_mode_1
        )
        ])
        self.assertEqual(discrete_event_simulation.context.c_events_set, expected_c_events_set)

    def test__execute_b_event__inspection(self):
        inspection_name = "long_shut_down"
        inspection_period = 4
        inspection_duration = 2

        phase = Phase("phase_test", DeterministicLaw([10]), True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test__execute_b_event",
                                     DeterministicLaw([10]),
                                     DeterministicLaw([10]),
                                     Failure(FailureType.DETECTABLE),
                                     [phase],
                                     Inspection(inspection_name, inspection_period, inspection_duration),
                                     [phase],
                                     None,
                                     'AFTER_REPAIR',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())


        context = Context(root_component, phase_manager, rca_manager, set(), set())

        inspection_event = StartInspectionEvent(0, context, _basic, failure_mode_1)

        discrete_event_simulation = DiscreteEventSimulation(0, 3, context)

        discrete_event_simulation._execute_b_event(inspection_event)

        self.assertEqual(_basic.status, Status.INSPECTION)

        expected_c_events_set = set([OrderEndInspectionEvent(
            priority=CEventPriority.ORDER_END_INSPECTION_EVENT,
            context=context,
            component=_basic,
            event=inspection_event,
            failure_mode=failure_mode_1
        )
        ])

        self.assertEqual(discrete_event_simulation.context.c_events_set, expected_c_events_set)

    def test__execute_b_event__blind_failure(self):
        phase = Phase("phase_test", DeterministicLaw([10]), True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test__execute_b_event",
                                     DeterministicLaw([1]),
                                     DeterministicLaw([1]),
                                     Failure(FailureType.BLIND),
                                     [phase],
                                     None,
                                     [phase],
                                     None,
                                     'NEVER',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager)

        blind_failure_event = BlindFailureEvent(0, context, _basic, failure_mode_1)

        discrete_event_simulation = DiscreteEventSimulation(0, 3, context)

        discrete_event_simulation._execute_b_event(blind_failure_event)

        self.assertEqual(_basic.status, Status.BLIND_FAILED)
        self.assertEqual(discrete_event_simulation.context.c_events_set, set())

    def test__execute_c_events(self):
        phase = Phase("phase_test", DeterministicLaw(10), True)
        phase.set_next_phase(PhaseManager.NO_PHASE)

        failure_mode_1 = FailureMode("test__execute_c_event",
                                     DeterministicLaw([1]),
                                     DeterministicLaw([1]),
                                     Failure(FailureType.DETECTABLE),
                                     [phase],
                                     None,
                                     [phase],
                                     None,
                                     'AFTER_REPAIR',
                                     [phase])

        phase_manager = PhaseManager([phase],set())

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], failure_mode_1, [], [])
        root_component.add_children_list([_basic])

        rca_manager = RootCauseAnalysisManager(set(), root_component.to_set())
        context = Context(root_component, phase_manager, rca_manager, set(), set())

        order_repair_event = OrderRepairEvent(priority=0,
                                              context=context,
                                              component=_basic,
                                              event=DetectableFailureEvent(0, context, _basic, failure_mode_1),
                                              failure_mode=failure_mode_1)

        discrete_event_simulation = DiscreteEventSimulation(0, 3, context, set(), set())
        discrete_event_simulation.context.c_events_set = set([order_repair_event])

        discrete_event_simulation._execute_c_events()

        expected_b_event_set = set([StartRepairingEvent(0, context, _basic, order_repair_event),
                                    EndRepairingEvent(1, context, _basic, order_repair_event, failure_mode_1)])

        self.assertEqual(expected_b_event_set, discrete_event_simulation.context.b_events_set)
        self.assertFalse(discrete_event_simulation.context.c_events_set)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer, Zip


class SensitivityAnalysisStrategyError(Exception):
    pass


class ExplorationStrategyFactory:
    """
    Factory class for the Exploration Strategy
    """

    @staticmethod
    def build(parameter_name,
              values,
              strategy_str):
        """
        Given a strategy_str it creates the corresponding ExplorationStrategy.
        Note: strategy_name is case UNsensitive.
        :param parameter_name name of the parameter onto which the strategy will be applied
        :param values the values to apply
        :param strategy_str the string representation of the strategy to apply.
        :raise SensitivityAnalysisStrategyException when the given strategy_str is not applicable.
        """
        if strategy_str.upper() == "INNER":
            return Inner(parameter_name, values)
        elif strategy_str.upper() == "OUTER":
            return Outer(parameter_name, values)
        elif strategy_str.upper() == "ZIP":
            return Zip(parameter_name, values)
        else:
            message_exception = f"{strategy_str} wrong strategy in sensitivity analysis for line"
            import logging
            logging.exception(message_exception)
            raise SensitivityAnalysisStrategyError(message_exception)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.results.result_record_entry import ResultRecordEntry


class ResultRecordEntryComponent(ResultRecordEntry):
    """
    This method defines an Identifier of the record entry.
    This is useful, in case we want to regroup the entry based on this identifier. {cf SimulationResults}
    That particular class is dedicated to entries of components.
    """
    __slots__ = 'component', 'status', 'phase', 'description'

    def __init__(self,
                 component,
                 status,
                 phase,
                 description: str):
        self.component = component
        self.status = status
        self.phase = phase
        self.description = description

    def __hash__(self):
        return hash((self.component, self.status, self.phase, self.description))

    def __eq__(self, other) -> bool:
        return self.component == other.component and \
               self.phase == other.phase and \
               self.status == other.status and \
               self.description == other.description

    def __str__(self) -> str:
        return f"ResultRecordEntryComponent: {self.component.name}__{self.component.local_id}__{self.component.global_id}"

    def identifier(self) -> str:
        """
        Function returning an identifier used to filter records
        """
        return f"{self.component.name}_{self.component.local_id}_{self.component.global_id}"



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure import FailureType, Failure
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.discrete_event_simulation.b_event_generator import BEventGenerator
from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import NextPhaseEvent
from availsim4core.src.timeline.record_component import RecordComponent


class test_Basic(unittest.TestCase):

    def test__eq__(self):
        basic_1 = Basic(1, "test_compound_1", 0, [], None, [], [])
        basic_2 = Basic(1, "test_compound_1", 0, [], None, [], [])

        self.assertEqual(basic_1, basic_2)

    def test_to_set(self):
        basic = Basic(1, "dummyName", 0, None, [None, None], [], [])
        result = basic.to_set()
        self.assertEqual(result, set([basic]))

    def test_get_failure_events(self):
        # failures modes used to define a Basic
        phase_set = set()
        rca_set = set()
        first_failure_mode = FailureMode("first failure mode",
                                         ProbabilityLaw("dummy", 1.0, False),
                                         ProbabilityLaw("dummy", 2.0, False),
                                         Failure(FailureType.DETECTABLE),
                                         [],
                                         None,
                                         phase_set,
                                         None,
                                         'AFTER_REPAIR',
                                         [])

        basic = Basic(1, "dummyName", 0, [], first_failure_mode, [], [])
        context = Context(basic, PhaseManager(phase_set, set()), rca_set)

        # mocking the get_random_value inside the probability law
        # such mocking is necessary as without it, the random value returned by "get_random_value" can not be predicted
        # the test would always fail
        with patch.object(ProbabilityLaw, "get_random_value", return_value=10.):

            # populate list of B events
            b_events_set = BEventGenerator.generate_first_events(0, context)
            c_events_set = CEventGenerator.generate_first_events(0, context)

            b_events_from_c_events_set = set()
            for c_event in c_events_set:
                if c_event.is_condition_valid():
                    b_events_from_c_events_set.update(c_event.generate_b_events(0))
            b_events_set.update(b_events_from_c_events_set)

        # generating expected events
        import numpy as np
        phase_event = NextPhaseEvent(np.inf, context)
        first_failure_event = DetectableFailureEvent(10.,
                                                     context,
                                                     basic,
                                                     first_failure_mode)

        # testing the equality between result and expected results
        self.assertEqual(b_events_set, set([first_failure_event, phase_event]))

    def test_update_status(self):
        context = Context(None, None, None)
        basic = Basic(1, "test_basic_1", 0, [], None, [], [])
        status = Status.FAILED
        timestamp = 10
        phase = PhaseManager.NO_PHASE
        desc = "test"
        result = basic.update_status(status, timestamp, phase, desc, context)
        expected_result = [RecordComponent(basic, status, timestamp, phase, desc)]
        self.assertEqual(result, expected_result)
        self.assertEqual(basic.status, status)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List


class ProbabilityLaw:
    """
    Distributive Function which is used to evaluate the failure and repair times.
    """
    __slots__ = 'name', 'parameters', 'is_failure_on_demand'

    def __init__(self,
                 name: str,
                 parameters: List[float],
                 is_failure_on_demand):
        self.name = name
        self.parameters = parameters
        self.is_failure_on_demand = is_failure_on_demand

    def __str__(self):
        return f"{self.name} -> {self.parameters}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.name == other.name and \
               self.parameters == other.parameters

    def get_random_value(self):
        pass

    def get_quantile_value(self, quantile):
        pass

    def get_mean_value(self):
        pass

    def get_variance_value(self):
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
import pandas

class XLSXReaderError(Exception):
    pass
class XLSXReaderEmptyStringError(Exception):
    pass

class XLSXReader:

    pragma_for_expression = '#'

    @staticmethod
    def read(file_path):
        """
        Read the XLSX given file with Panda and returns it as a Panda data Frame.
        -> dictionary: sheetsName
            --> dictionary: Lines Number
                --> dictionary: columnName
                        -> Value
        :param file_path: .xlsx file path
        :return: the Panda data frame of the given file.
        """
        data = pandas.read_excel(file_path, sheet_name=None, engine='openpyxl', keep_default_na=False)

        dictionary = {}
        for key in data.keys():
            dictionary[key] = data[key].to_dict('index')
        return dictionary

    @staticmethod
    def clean_cell_str_else_raise(entry_str: str, exception_message_hint: str = ""):
        """
        Extracts the unique word within a string, removing any space (" ") and forcing to upper
        :param entry_str
        :param exception_message_hint: hint about where does the string comes from in order to print a debug information
        """
        ret = entry_str.replace(" ", "").replace("'", "").replace('"', '').replace('“', '').replace('”', '').upper()
        if ret != '':
            return ret
        else:
            message_exception = f"Empty string in one cell, here is an exception_message_hint, maybe: {exception_message_hint}"
            logging.exception(message_exception)
            raise XLSXReaderEmptyStringError(message_exception)

    @staticmethod
    def extract_from_str_list(cell_entry_str: str, exception_message_hint: str = ""):
        """
        Extracts elements from a given list in the format '[1., 2., 3.]'
        :param cell_entry_str is a string representing either a list of float or a list of string.
        :param exception_message_hint: hint about where does the string comes from in order to print a debug information
        """

        # if the string start with a '#', then it's some expression to evaluate
        if cell_entry_str[0] == XLSXReader.pragma_for_expression:
            import numpy  # for a possible execution of python code in the "eval" of function extract_from_str_list
            output_list = eval(cell_entry_str[1:])
            if isinstance(output_list,(numpy.ndarray,range)):
                output_list=list(output_list)
            if not isinstance(output_list,list):
                message_exception = f"A string containing {XLSXReader.pragma_for_expression} as first character has been" \
                                    f" evaluated to extract a list out of it but the attempt failed. The content of the" \
                                    f" string is {cell_entry_str}, the variable evaluated is {output_list} of type {type(output_list)}"
                import logging
                logging.exception(message_exception)
                raise XLSXReaderError(message_exception)

        else:

            # if the string is a list with explicit brackets, we remove them
            if cell_entry_str[0] == "[" and cell_entry_str[-1] == "]":
                cell_entry_str = cell_entry_str[1:-1]

            # any space, quote, double quote are removed; all those symbols are likely artefacts of lists
            clean_list = XLSXReader.clean_cell_str_else_raise(cell_entry_str,exception_message_hint=exception_message_hint).split(",")

            try:
                # if the list is only composed of numeric, we return floats
                output_list = [float(i) for i in clean_list]
            except ValueError:
                # else we redo the analysis, trying to spot any list within a string
                output_list = [XLSXReader.transform_str_of_list_in_list(i) for i in clean_list]

        return output_list

    @staticmethod
    def transform_str_of_list_in_list(str_in):
        # test to check if a string contains a list
        if str_in[0] == "[" and str_in[-1] == "]":
            return XLSXReader.extract_from_str_list(str_in)
        else:
            return str_in


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class SystemModifierCombination:
    """
    SystemTemplate modifier Combination describes a new system based on the Sensitivity Analysis file.
    It defines the Combination list which needs to be applied on the initial SystemTemplate
    in order to get all the analysis.
    """

    def __init__(self, system_modifier_list):
        self.system_modifier_list = system_modifier_list

    def __str__(self):
        return str(self.system_modifier_list)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.system_modifier_list == other.system_modifier_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import itertools
import logging

from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_inner_strategy import \
    SystemModifierCombinationInnerStrategy
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_outer_strategy import \
    SystemModifierCombinationOuterStrategy
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_zip_strategy import \
    SystemModifierCombinationZipStrategy


class SystemModifierCombinationStrategyDirector:
    """
    Director Strategy class which defines the sensitivity strategies to apply.
    """

    @staticmethod
    def execute(exploration_strategy_list):
        """
        Defines the strategy to apply on the given exploration strategy list
        and returns the corresponding {SystemModifierCombination}
        :param exploration_strategy_list the list of {ExplorationStrategy}
        """
        system_modifier_combination_inner_list = SystemModifierCombinationInnerStrategy.execute(exploration_strategy_list)
        system_modifier_combination_zip_list = SystemModifierCombinationZipStrategy.execute(exploration_strategy_list)
        system_modifier_combination_outer_list = SystemModifierCombinationOuterStrategy.execute(exploration_strategy_list)

        system_modifier_combination_list = []

        for values_tuple in itertools.product(*[_list for _list in [
            system_modifier_combination_inner_list,
            system_modifier_combination_zip_list,
            system_modifier_combination_outer_list
            ]
            if _list]):

            local_system_modifier_combination_list=[]
            for value in values_tuple:

                local_system_modifier_combination_list.extend(value.system_modifier_list)

            system_modifier_combination_list.append(SystemModifierCombination(local_system_modifier_combination_list))

        logging.debug(f"sensitivity modifier combination list : {system_modifier_combination_list}")
        return system_modifier_combination_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterExecutionMetrics(ResultExporter):
    """
    class used to export the results of a simulation into an output file
    This exporter is meant to provide a good "summary" of a simulation, not a light report like the exporter LightStudy
    """

    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        execution_metrics_dict = {"execution_time": execution_time}
        execution_metrics_dict['number_of_DES_simulations_executed'] = str(results.number_of_DES_simulations_executed)
        for key, statistics in results.execution_metrics_statistics.items():
            execution_metrics_dict[key + "_MEAN"] = statistics.mean
            execution_metrics_dict[key + "_MAX"] = statistics.max
            execution_metrics_dict[key + "_STD"] = statistics.std

        number_of_compound_components = 0
        number_of_basic_components = 0

        for component in context.root_component.to_set():
            if component.get_children():
                # if that component has children (then it is a Compound type)
                number_of_compound_components += 1
            else:
                number_of_basic_components += 1

        execution_metrics_dict["number_of_compound_components"] = number_of_compound_components
        execution_metrics_dict["number_of_basic_components"] = number_of_basic_components

        # exporting the execution time and other metric
        execution_metrics_dataframe = pandas.DataFrame(
            execution_metrics_dict,
            index=[0]) # TODO add the number of iteration? number of b events? number of components?
        execution_metrics_dataframe.to_excel(writer, sheet_name='RESULTS_EXECUTION_METRICS', index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import Dict

from availsim4core.src.reader.xlsx.xlsx_reader import XLSXReader
from availsim4core.src.simulation.monte_carlo import MonteCarlo


class SimulationMonteCarloColumn:
    MINIMUM_NUMBER_OF_SIMULATION = "MIN_NUMBER_OF_SIMULATION"
    MAXIMUM_NUMBER_OF_SIMULATION = "MAX_NUMBER_OF_SIMULATION"
    CONVERGENCE_MARGIN = "CONVERGENCE_MARGIN"
    MAXIMUM_EXECUTION_TIME = "MAX_EXECUTION_TIME"
    SEED = "SEED"
    DIAGNOSTICS = "DIAGNOSTICS"
    DURATION = "SIMULATION_DURATION"


class MonteCarloReader(XLSXReader):
    """
    Specific reader for the MonteCarlo simulation.
    """

    @classmethod
    def build(cls, simulation_parameters: Dict):
        return MonteCarlo(
            int(simulation_parameters[SimulationMonteCarloColumn.MINIMUM_NUMBER_OF_SIMULATION]),
            int(simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_NUMBER_OF_SIMULATION]),
            float(simulation_parameters[SimulationMonteCarloColumn.CONVERGENCE_MARGIN]),
            float(simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_EXECUTION_TIME]),
            int(simulation_parameters[SimulationMonteCarloColumn.SEED]),
            cls.extract_from_str_list(simulation_parameters[SimulationMonteCarloColumn.DIAGNOSTICS]),
            float(simulation_parameters[SimulationMonteCarloColumn.DURATION])
        )


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateReader


class test_template(unittest.TestCase):
    param_scenario = [
        ["./availsim4core/test/E2E/input/template/simulation.xlsx",
         "./availsim4core/test/E2E/input/template/system.xlsx"],
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file)

    def _runner_E2E(self,
                    simulation_file,
                    system_file):

        output_folder = "./availsim4core/test/E2E/output/template/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, None, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")

        # read both file and then compare?
        input_system_template = SystemTemplateReader.generate_system_template(system_file,None)
        output_system_template = SystemTemplateReader.generate_system_template(result_simulation_file_list[0],None)

        self.assertEqual(input_system_template,output_system_template)

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.component import Component

from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.rca.rca_manager import RootCauseAnalysisManager
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.timeline import Timeline


class Context:
    """
    This class represents the overall context state during a discrete event simulation.
    """
    __slots__ = 'root_component', 'phase_manager', 'timeline_record', 'rca_manager', 'b_events_set', 'c_events_set', 'absolute_simulation_time'

    def __init__(self,
                 root_component: Component,
                 phase_manager: PhaseManager, 
                 rca_manager: RootCauseAnalysisManager,
                 b_events_set: Set = set(),
                 c_events_set: Set = set()):
        self.root_component = root_component
        self.phase_manager = phase_manager
        self.timeline_record = Timeline()
        self.rca_manager = rca_manager
        self.b_events_set = b_events_set
        self.c_events_set = c_events_set
        self.absolute_simulation_time = 0

    def __eq__(self, other):
        return self.root_component == other.root_component and \
                self.phase_manager == other.phase_manager and \
                self.timeline_record == other.timeline_record and \
                self.rca_manager == other.rca_manager

    def __str__(self):
        return (f"Context : root_component name={self.root_component.name}, "
                f"current_phase={self.phase_manager.current_phase}")

    def add_record_of_each_component(self,
                                     absolute_time: float,
                                     description: str):
        """
        Function adding a record for each component.
        """
        records = []
        for component in self.root_component.to_set():
            records.append(
                RecordComponent(component,
                                component.status,
                                absolute_time,
                                self.phase_manager.current_phase,
                                description)
            )
        return records



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import numpy
from numpy.random import weibull
from scipy.special import gamma
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw


class WeibullLaw(ProbabilityLaw):
    """
    Class used to generate values according to the Weibull law
    """

    def __init__(self,
                 parameters):
        super().__init__(self.__class__.__name__,
                         parameters,
                         is_failure_on_demand=False)

    def __eq__(self, other):
        if not isinstance(other, WeibullLaw):
            return NotImplemented
        return super().__eq__(other)

    def get_random_value(self):
        return self.parameters[0]*weibull(self.parameters[1])

    def get_quantile_value(self, quantile):
        return self.parameters[0]*(-numpy.log(1-quantile)**(1/self.parameters[1]))

    def get_mean_value(self):
        return self.parameters[0]*gamma(1+1/self.parameters[1])

    def get_variance_value(self):
        return self.parameters[0] ** 2 * (
            gamma(1 + 2 / self.parameters[1]) - gamma(1 + 1 / self.parameters[1]) ** 2
        )


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.record_generator import RecordGenerator
from availsim4core.src.timeline.record_phase import RecordPhase
from availsim4core.src.timeline.timeline import Timeline


class test_Timeline(unittest.TestCase):

    def test_extract_record_from_types(self):
        basic_1 = Basic(1, "test_basic_1", 0, [], [], [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], [], [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)

        record_phase = RecordPhase(PhaseManager.NO_PHASE,
                                   0, RecordGenerator.DESCRIPTION_INIT)

        record_list = [record_phase,
                       RecordComponent(basic_1,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(basic_2,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT),
                       RecordComponent(compound,
                                       Status.RUNNING,
                                       0,
                                       PhaseManager.NO_PHASE,
                                       RecordGenerator.DESCRIPTION_INIT)
                       ]

        timeline = Timeline()
        timeline.record_list = record_list
        result = timeline.extract_record_from_types(RecordPhase)
        expected_result = [record_phase]

        self.assertEqual(result, expected_result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
import os
from datetime import datetime
from typing import List

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.exporter.xlsx_analysis_exporter import XLSXAnalysisExporter


class HTCondorRunner:
    """
    class used to run simulations on a cluster interfaced by HTCondor
    """

    # Note : This does not work due to permissions issues.
    #  Did not find a quick working solution, let's postpone it as the solution below is enough so far...
    #
    # @staticmethod
    # def run(root_output_folder, path_simulation, path_system, system_template):
    #     date_time_str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    #     output_folder = root_output_folder + "run___" + date_time_str + "___" + str(system_template.id) + "/"
    #     os.makedirs(output_folder, exist_ok=True)
    #
    #     hostname_job = htcondor.Submit({
    #         "executable": "python3 availsim4.py",  # the program to run on the execute node
    #         "--system": path_system,
    #         "--simulation": path_simulation,
    #         "--output_folder": output_folder,
    #         "output": "out.txt",  # anything the job prints to standard output will end up in this file
    #         "error": "err.txt",  # anything the job prints to standard error will end up in this file
    #         "log": "log.txt",  # this file will contain a record of what happened to the job
    #         "request_cpus": "1",  # how many CPU cores we want
    #     })
    #
    #     print(hostname_job)

    @staticmethod
    def create_one_analysis_folder(root_output_folder:str,
                                   pipinstall_folder: str,
                                   date_time_str: str,
                                   analysis: Analysis,
                                   HTCondor_extra_argument: str):
        """
        :param root_output_folder: folder in which the analysis' folder will be created
        :param pipinstall_folder: folder where the requirements are installed
        :param date_time_str: unique identifier used to name the new folders
        :param analysis: the analysis to perform in the folder
        :param HTCondor_extra_argument: any extra argument used when submitting the job
        :return: nothing, the function just create folders and files
        """

        # creating a unique subdirectory containing on of the simulation of the sensitivity analysis
        output_folder = root_output_folder + "run___" + date_time_str + "___" + str(analysis.id) + "/"
        os.makedirs(output_folder,
                    exist_ok=True)
        logging.debug(f"Creating HTCondor output folder {output_folder}")

        # copying simulation file
        simulation_filename = output_folder + "simulation.xlsx"
        system_filename = output_folder + "system.xlsx"

        writer_system = pandas.ExcelWriter(system_filename, engine='xlsxwriter')
        XLSXAnalysisExporter.export_system(writer_system,
                                           analysis.system_template)
        writer_system.save()

        writer_simulation = pandas.ExcelWriter(simulation_filename, engine='xlsxwriter')
        XLSXAnalysisExporter.export_simulation(writer_simulation,
                                               analysis.simulation)
        writer_simulation.save()

        code_directory = os.getcwd()
        bash_filename = output_folder + "run.sh"
        with open(bash_filename, 'w') as file:
            try:
                file.write(
                    f"#!/usr/bin/env bash\n"
                    f"source /afs/cern.ch/eng/tracking-tools/python_installations/activate_default_python\n"
                    f"export PYTHONPATH={pipinstall_folder}\n"
                    f"export PYTHONNOUSERSITE=1\n"
                    f"python3.7 {code_directory}/availsim4.py"
                    f" --system {system_filename}"
                    f" --simulation {simulation_filename}"
                    f" --output_folder {output_folder}\n"
                    f"tar -czvf out.tar.gz out.txt\n"
                    f"rm -rf out.txt"
                )
            except IOError as e:
                print(f"I/O error({e.errno}): {e.strerror}")

        os.system("chmod +x " + bash_filename)

        runtime = int(analysis.simulation.maximum_execution_time * 1.1 + 10)

        # creating submission file for a unique job
        sub_filename = output_folder + "sub.sub"
        with open(sub_filename, 'w') as file:
            try:
                content_of_sub = (f"executable = {bash_filename}\n"
                                  f"output = {output_folder + 'out.txt'}\n"
                                  f"error = {output_folder + 'err.txt'}\n"
                                  f"log = {output_folder + 'log.txt'}\n"
                                  "RequestCpus = 1\n"
                                  f"+MaxRuntime = {runtime}\n"
                                  f"{HTCondor_extra_argument}\n"
                                  f"queue")
                file.write(content_of_sub)
            except IOError as e:
                logging.info(f"I/O error({e.errno}): {e.strerror}")

    @staticmethod
    def run(root_output_folder: str,
            analysis_list: List[Analysis],
            HTCondor_extra_argument: str):
        """
        The runner of HTCondor is taking care of running jobs on HTCondor. This is done in 2 steps:
        1) checking that the requirements are installed in the pipinstall folder
        2) creating the directory / input file / etc of each analysis one wants to perform
        3) submitting all the analyses in one job, requesting as many cores as analyses
        """

        date_time_str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        # creating a pipinstall folder for the analysis
        pipinstall_folder = root_output_folder + "pipinstall"
        if os.path.isdir(pipinstall_folder):
            logging.info("pipinstall folder already exists, no need to create it")
        else:
            logging.info("pipinstall folder does not already exist, installation of requirements")
            code_directory = os.getcwd()
            os.makedirs(pipinstall_folder, exist_ok=True)
            command = f"pip install --target={pipinstall_folder} -r {code_directory}/requirements.txt"
            os.system(command)

        # creating one folder per analysis
        for analysis in analysis_list:
            HTCondorRunner.create_one_analysis_folder(
                root_output_folder,
                pipinstall_folder,
                date_time_str,
                analysis,
                HTCondor_extra_argument)

        # creating the master bash file
        master_bash_filename = root_output_folder + "run.sh"
        with open(master_bash_filename, 'w') as file:
            try:
                file.write("#!/usr/bin/env bash" + "\n")
                file.write("cd " + root_output_folder + "*__$1" + "\n")
                file.write("source ./run.sh" + "\n")
            except IOError as e:
                print(f"I/O error({e.errno}): {e.strerror}")
        os.system("chmod +x " + master_bash_filename)

        # creating the master sub file

        master_sub_filename = root_output_folder + "sub.sub"
        runtime = int(analysis_list[0].simulation.maximum_execution_time * 1.1 + 10)

        output_folder = root_output_folder + "/run___" + date_time_str + "___" + "$(ProcId)/"

        with open(master_sub_filename, 'w') as file:
            try:
                content_of_sub = (f"executable = {master_bash_filename}\n"
                                  f"arguments = $(ProcId)\n"
                                  f"output = {output_folder+'out.txt'}\n"
                                  f"error = {output_folder+'err.txt'}\n"
                                  f"log = {output_folder+'log.txt'}\n"
                                  "RequestCpus = 1\n"
                                  f"+MaxRuntime = {runtime}\n"
                                  f"{HTCondor_extra_argument}\n"
                                  f"queue {len(analysis_list)}")
                file.write(content_of_sub)

            except IOError as e:
                logging.info(f"I/O error({e.errno}): {e.strerror}")

        # submitting a unique request to HTCondor
        command = "cd " + root_output_folder + "; condor_submit " + master_sub_filename
        logging.debug(f"Executing HTCondor command {command}")
        os.system(command)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterComponentTreeExtended(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        def component_tree_extended_export(component: Component, level=0):
            """
            Function recursively exploring the system in order to build a graphical representation of the system.
            This "extended" version of the tree exporter provides a detailed output. A simple version exists too.
            The function is called on the root component. Then if a component has children, it is called again for its
            children.
            """

            # Variable used to get the logic between the parent of the current component and the current component.
            # It is possible that a component has several parents if it's a "shared child".
            dict_of_children_logic = {}
            for parent in component.get_parents():
                dict_of_children_logic[parent.__repr__()] = parent.children_logic

            if level == 0:
                # if level ==0, then the current component is the root component, it has no parent,
                # only a "light" string is used to described the component
                ret = (f";" * (level) +
                       f"{component.name}_{component.local_id}_{component.global_id}; ROOT COMPONENT HAS NO PARENT \n")
            else:
                # else, the current component has some parent(s),
                # a more detailed string is used, with logic applied by its parent(s)
                ret = (f";" * (level - 1) + f"logic = {dict_of_children_logic};"
                                            f"{component.name}_{component.local_id}_{component.global_id}; parents = {component.get_parents()} \n")

            if component.get_children():
                # if that component has children (then it is a Compound type), apply the same function to the children
                for child in component._children:
                    ret += component_tree_extended_export(child,level + 1)
            else:
                # else, the current component is a basic and has failure mode(s), those a printed
                ret += (f";" * (level + 1) +
                        f"failure mode = {component.failure_mode}\n")

            return ret

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        tree_string = component_tree_extended_export(context.root_component)

        tree_dataframe = pandas.DataFrame([x.split(';') for x in tree_string.split('\n')])

        tree_dataframe.to_excel(writer, sheet_name='RESULTS_COMPONENT_TREE_EXTENDED', index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.end_repairing_event import EndRepairingEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.start_repairing_event import \
    StartRepairingEvent
from availsim4core.src.discrete_event_simulation.event.c_event.component_c_event import ComponentCEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class OrderRepairEvent(ComponentCEvent):
    """
    class used to regenerate all InspectionEvents at once, once they have all been executed
    C events are generated which themselves generate B events
    """
    __slots__ = 'event', 'failure_mode'

    def __init__(self,
                 priority: int,
                 context: Context,
                 component: Component,
                 event: Event,
                 failure_mode: FailureMode):
        super().__init__(priority, context, component)
        self.event = event
        self.failure_mode = failure_mode

    def __eq__(self, other):
        if not isinstance(self, OrderRepairEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event and \
               self.failure_mode == other.failure_mode

    def __hash__(self):
        return hash((type(self), self.priority, self.component, self.event, self.failure_mode))

    def __str__(self):
        return f"OrderRepairEvent:: " \
               f"priority:{self.priority} - " \
               f"component:{self.component.name} - " \
               f"failure_mode:{self.failure_mode.name}"

    def generate_b_events(self, absolute_simulation_time):
        start_repairing_event = StartRepairingEvent(absolute_occurrence_time=absolute_simulation_time,
                                                    context=self.context,
                                                    basic=self.component,
                                                    event=self)

        end_repairing_event = EndRepairingEvent(
            absolute_occurrence_time=absolute_simulation_time + self.failure_mode.repair_law.get_random_value(),
            context=self.context,
            basic=self.component,
            event=self,
            failure_mode=self.failure_mode)

        return {start_repairing_event, end_repairing_event}



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set

from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.event import Event


class MinimalReplaceableUnitEndRepairingEvent(BasicBEvent):
    """
    This event defines the End of a reparation for a Minimal Replaceable Unit.
    """
    __slots__ = 'event', 'mru_trigger'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 event: Event,
                 mru_trigger: MinimalReplaceableUnit):
        super().__init__(absolute_occurrence_time, context, basic,
                         BEventPriority.MRU_END_REPAIRING_EVENT)
        self.event = event
        self.mru_trigger = mru_trigger

    def __hash__(self):
        return hash(
            (type(self), self.absolute_occurrence_time, self.priority, self.basic, self.event, self.mru_trigger))

    def __eq__(self, other):
        if not isinstance(other, MinimalReplaceableUnitEndRepairingEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event and \
               self.mru_trigger == other.mru_trigger

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"mru trigger ::{self.mru_trigger}, "
                f"at t ::{self.absolute_occurrence_time} \n")

    def execute(self):
        return self.basic.update_status(Status.RUNNING,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,  # or phase of the failure mode?
                                        f"mru ({self.event.mru.name}) REPAIRED",
                                        self.context)

    def generate_c_event(self, **kwargs) -> Set[CEvent]:
        from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
        from availsim4core.src.discrete_event_simulation.event.c_event. \
            repair_event.minimal_replaceable_unit_order_repair_event import MinimalReplaceableUnitOrderRepairEvent

        set_of_c_events = set()

        set_of_c_events.update(
            CEventGenerator.get_c_events_for_failures(
                self.absolute_occurrence_time,
                self.context,
                self.basic
            )
        )

        set_of_c_events.add(
            MinimalReplaceableUnitOrderRepairEvent(
                priority=CEventPriority.MRU_ORDER_REPAIR_EVENT,
                context=self.context,
                component=self.event.component,
                mru=self.mru_trigger
            )
        )

        return set_of_c_events


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List, Set, Tuple

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEvent, BEventPriority
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event import detectable_failure_event, \
    blind_failure_event
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event import FailureEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_event import \
    OrderNextPhaseEvent
from availsim4core.src.discrete_event_simulation.event.c_event.postpone_c_event import PostponeCEvent
from availsim4core.src.timeline.record_phase import RecordPhase


class NextPhaseEvent(BEvent):
    """
    Class handling the change of phases
    """

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 priority=BEventPriority.NEXT_PHASE_EVENT):
        super().__init__(absolute_occurrence_time,
                         context,
                         priority)

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority))

    def __eq__(self, other):
        if not isinstance(other, NextPhaseEvent):
            return NotImplemented
        return super().__eq__(other)

    def __str__(self):
        return (f"type ::{type(self)} "
                f"at t ::{self.absolute_occurrence_time} \n")

    @staticmethod
    def _b_events_to_be_cleaned() -> List:
        return [detectable_failure_event.DetectableFailureEvent,
                blind_failure_event.BlindFailureEvent,
                NextPhaseEvent]

    @staticmethod
    def _c_events_to_be_cleaned() -> List:
        from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_if_failure_event import \
            OrderNextPhaseIfFailureEvent
        return [OrderNextPhaseIfFailureEvent,
                OrderNextPhaseEvent]

    def update_b_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """
        b_event_to_remove_set = set()
        b_event_to_remove_and_not_to_postpone = set()
        previous_phase_record = self.context.timeline_record._get_previous_record_of_type(RecordPhase)
        if previous_phase_record is not None:
            for b_event in event_set:
                if isinstance(b_event,NextPhaseEvent):
                    b_event_to_remove_and_not_to_postpone.add(b_event)
                elif isinstance(b_event, FailureEvent):
                    if previous_phase_record.phase not in b_event.failure_mode.phase_set:
                        b_event_to_remove_set.add(b_event)
        return event_set - b_event_to_remove_set - b_event_to_remove_and_not_to_postpone, b_event_to_remove_set

    def execute(self):
        records = [self.context.phase_manager.go_to_phase(
            self.absolute_occurrence_time,
            self.context.phase_manager.current_phase.next,
            self.context.phase_manager.DEFAULT_PHASE_DESCRIPTION
        )]
        records.extend(self.context.add_record_of_each_component(self.absolute_occurrence_time,
                                                                 self.context.phase_manager.UPDATE_FOLLOWING_A_PHASE_CHANGE))
        return records

    def generate_c_event(self, **kwargs) -> Set:

        c_event_set: Set[CEvent] = {
            OrderNextPhaseEvent(
                priority=CEventPriority.ORDER_NEXT_PHASE_EVENT,
                context=self.context
            )
        }

        event_set = kwargs.pop('b_event_removed_set', set())
        previous_phase_record = self.context.timeline_record._get_previous_record_of_type(RecordPhase)
        if previous_phase_record is None:
            return c_event_set

        for b_postpone_event in event_set:

            if not isinstance(b_postpone_event,NextPhaseEvent):

                # Search for the previous record where the concerned basic component was running.
                # Either it entered the phase as RUNNING and we postpone the whole duration of the phase
                # Or it entered the phase as something else than RUNNING, but still it has been changed to RUNNING
                # otherwise no failure event would need to be postponed
                previous_record = self.context.timeline_record._get_previous_record_of_basic_in_status(
                    b_postpone_event.basic,
                    Status.RUNNING)

                c_event_set.add(
                    PostponeCEvent(
                        priority=CEventPriority.ORDER_POSTPONE_C_EVENT,
                        context=self.context,
                        postpone_duration=self.absolute_occurrence_time - previous_record.timestamp,
                        b_event=b_postpone_event
                    )
                )

        return c_event_set



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit


class Compound(Component):
    """
    Node of the Component tree.
    Compound is a composite thus it includes a Children list of Component.
    """
    __slots__ = '_children', 'children_logic'

    def __init__(self,
                 global_id: int,
                 name: str,
                 local_id: int,
                 parents: List[Compound],
                 children_logic: ChildrenLogic,
                 list_of_mru_trigger: List[MinimalReplaceableUnit]):
        super().__init__(global_id, name, local_id, parents, list_of_mru_trigger)
        self._children: List[Component] = []
        self.children_logic = children_logic

    def __str__(self, level=0):
        ret = ("\t" * level +
               f"{self.name}_{self.local_id}"
               f" :uniq_id:{self.global_id}"
               f" :parents: {self.get_parents()}"
               f" :status: {self.status.name}"
               f" :mru: {self.list_of_mru_trigger}\n")

        for child in self._children:
            ret += child.__str__(level + 1)
        return ret

    def __repr__(self):
        return f"{self.name}_{self.local_id}_{self.global_id}"

    def get_children(self):
        return self._children

    def add_children_list(self, component_list) -> None:
        """
        Add a list of Component to the Children list of component.
        :param component_list: the list of component Children to add.
        """
        self._children.extend(component_list)

    def remove(self, component: Component) -> None:
        self._children.remove(component)
        component._parents = []

    def update_status(self,
                      timestamp: float,
                      phase: Phase,
                      description: str,
                      context: Context):
        new_status = self.evaluate_status(context)
        return self._update_status(new_status, timestamp, phase, description, context)

    def evaluate_status(self, context: Context) -> Status:
        return self.children_logic.evaluate(self._children, context)

    def to_set(self):
        set_of_components = set([self])
        for child in self._children:
            set_of_components.update(child.to_set())
        return set_of_components



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_inner_strategy import \
    SystemModifierCombinationInnerStrategy


class test_SystemModifierCombinationInnerStrategy(unittest.TestCase):

    def test_execute(self):
        exploration_strategy_list = [Outer("outer1", [1., 2.]),
                                     Inner("inner1", [3., 4.]),
                                     Inner("inner2", [5., 6.])]

        system_modifier_list_inner_1 = [SystemModifier("inner1", 3.)]
        system_modifier_list_inner_2 = [SystemModifier("inner1", 4.)]
        system_modifier_list_inner_3 = [SystemModifier("inner2", 5.)]
        system_modifier_list_inner_4 = [SystemModifier("inner2", 6.)]
        expected_results = [SystemModifierCombination(system_modifier_list_inner_1),
                            SystemModifierCombination(system_modifier_list_inner_2),
                            SystemModifierCombination(system_modifier_list_inner_3),
                            SystemModifierCombination(system_modifier_list_inner_4)]

        results = SystemModifierCombinationInnerStrategy.execute(exploration_strategy_list)

        self.assertEqual(expected_results, results)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

"""
Module used to define sanity checks which are tested at each execution to avoid simulating invalid input files.
This is both used to be sure that the user did not forget / misunderstand something and also to avoid starting
simulations which after many hours will eventually stop because of an inconsistency in the input file or a bug caused
by a combination of parameters which is not allowed.
"""

import logging

from typing import List

from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.context.system.children_logic.required_component import RequiredComponent
from availsim4core.src.context.system.children_logic.tolerated_fault import ToleratedFault
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.system_template import SystemTemplate

from availsim4core.src.context.system.system_utils import SystemUtils


class ChildrenLogicSanityCheckError(Exception):
    pass


def _compare_and_log_diff(tree_list: List[str],
                         system_template_list: List[str],
                         description: str) -> bool:
    """
    Function taking two lists of strings, comparing them and logging the results using a description string.
    This function is specialized to compare names defined in the input file and names found in the tree simulated
    """

    matching = True

    tree_set = set(tree_list)
    template_set = set(system_template_list)

    tree_minus_input = tree_set - template_set
    if tree_minus_input:
        matching = False
        logging.warning(f"{description} present in the tree but not in the system template:{tree_minus_input}")

    input_minus_tree = template_set - tree_set
    if input_minus_tree:
        matching = False
        logging.warning(f"{description} present in the system template but not in the tree:{input_minus_tree}")

    return matching

def export_content_of_tree(root: Component) -> dict:
    """
    Function exploring a tree and extracting the information present in it
    :param root: component being the root of the tree to study
    :return: dict of list of names present in the tree
    """

    list_of_component_names_in_the_tree = []
    list_of_failure_mode_names_in_the_tree = []
    list_of_inspection_names_in_the_tree = []
    list_of_phase_names_in_the_tree = []
    list_of_mru_names_in_the_tree = []

    for component in root.to_set():
        list_of_component_names_in_the_tree.append(component.name)
        if not component.get_children():
            list_of_failure_mode_names_in_the_tree.append(component.failure_mode.name)
            for phase in component.failure_mode.phase_set:
                list_of_phase_names_in_the_tree.append(phase.name)
                list_of_phase_names_in_the_tree.append(phase.next.name)
                list_of_phase_names_in_the_tree.append(phase.next_phase_if_failure.name)
            if component.failure_mode.failure_mode_next_phase_if_failure is not None:
                list_of_phase_names_in_the_tree.append(component.failure_mode.failure_mode_next_phase_if_failure.name)
            if component.failure_mode.held_before_repair_phase_set is not set():
                list_of_phase_names_in_the_tree.extend(
                    [phase.name for phase in component.failure_mode.held_before_repair_phase_set])
            if component.failure_mode.held_after_repair_phase_set is not set():
                list_of_phase_names_in_the_tree.extend(
                    [phase.name for phase in component.failure_mode.held_after_repair_phase_set])
            if component.failure_mode.inspection is not None:
                list_of_inspection_names_in_the_tree.append(component.failure_mode.inspection.name)
            for MRU in component.list_of_mru_group:
                list_of_mru_names_in_the_tree.append(MRU.name)
        for MRU in component.list_of_mru_trigger:
            list_of_mru_names_in_the_tree.append(MRU.name)
            list_of_component_names_in_the_tree.extend(MRU.scope_common_ancestor)

    # TODO: SANITY CHECK: check phase jump

    return {"component names":list_of_component_names_in_the_tree,
            "failure mode names":list_of_failure_mode_names_in_the_tree,
            "inspection names":list_of_inspection_names_in_the_tree,
            "phase names":list_of_phase_names_in_the_tree,
            "mru names":list_of_mru_names_in_the_tree}

def export_content_of_system_template(system_template: SystemTemplate) -> dict:
    """
    Function exploring a system template (the object corresponding to the input file) and extracting the information
     present in it
    :param system_template:
    :return: dict of list of names present in the tree
    """

    list_of_component_names_in_the_system_template = []
    list_of_failure_mode_names_in_the_system_template = []
    list_of_inspection_names_in_the_system_template = []
    list_of_phase_names_in_the_system_template = []
    list_of_mru_names_in_the_system_template = []

    # architecture entries
    for architecture_entry in system_template.architecture_entry_list:
        list_of_component_names_in_the_system_template.append(architecture_entry.component_name)
        list_of_component_names_in_the_system_template.extend([SystemUtils.extract_name_of_function_from_string(entry)
                                                               for entry in architecture_entry.children_name_list])
        list_of_mru_names_in_the_system_template.extend(architecture_entry.in_mru_str_list)
        list_of_mru_names_in_the_system_template.extend(architecture_entry.trigger_mru_str_list)

    # failure mode assignments entries
    for failure_mode_assignments in system_template.failure_mode_assignments_list:
        list_of_component_names_in_the_system_template.append(failure_mode_assignments.component_name)
        list_of_failure_mode_names_in_the_system_template.append(failure_mode_assignments.failure_mode.name)

    # failure mode assignments entries
    for failure_mode in system_template.failure_mode_list:
        list_of_failure_mode_names_in_the_system_template.append(failure_mode.name)
        list_of_phase_names_in_the_system_template.extend([p.name for p in failure_mode.phase_set])
        list_of_phase_names_in_the_system_template.extend([p.name for p in failure_mode.held_after_repair_phase_set])
        list_of_phase_names_in_the_system_template.extend([p.name for p in failure_mode.held_before_repair_phase_set])
        if failure_mode.failure_mode_next_phase_if_failure is not None:
            list_of_phase_names_in_the_system_template.append(failure_mode.failure_mode_next_phase_if_failure.name)

    # MRU entries
    for mru in system_template.mru_list:
        list_of_mru_names_in_the_system_template.append(mru.name)
        list_of_component_names_in_the_system_template.extend(mru.scope_common_ancestor)

    # PHASES entries
    for phase in system_template.phase_set:
        list_of_phase_names_in_the_system_template.append(phase.name)
        list_of_phase_names_in_the_system_template.append(phase.next_phase_if_failure.name)

    # INSPECTION entries
    for inspection in system_template.inspection_list:
        list_of_inspection_names_in_the_system_template.append(inspection.name)

    return {"component names":list_of_component_names_in_the_system_template,
            "failure mode names":list_of_failure_mode_names_in_the_system_template,
            "inspection names":list_of_inspection_names_in_the_system_template,
            "phase names":list_of_phase_names_in_the_system_template,
            "mru names":list_of_mru_names_in_the_system_template}

def children_logic_validity(component: Component):
    """
    Recursive function parsing the tree to check the validity of the children logic, if it is not a custom one
    :param component: component which is currently studied
    :return: nothing, but raise an error if some inconsistency is spotted
    """

    if component.get_children():
        number_of_children = len(component.get_children())

        if isinstance(component.children_logic, RequiredComponent):
            if number_of_children < component.children_logic.minimum_number_of_required_component:
                message = f"{type(component.children_logic)} children logic of component {component.name} has less" \
                          f" children ({number_of_children}) than the minimum number of required components (" \
                          f"{component.children_logic.minimum_number_of_required_component}), thus the parent " \
                          f"component can never be in a RUNNING status."
                logging.exception(message)
                raise ChildrenLogicSanityCheckError(message)
            elif number_of_children == component.children_logic.minimum_number_of_required_component:
                message = f"{type(component.children_logic)} children logic of component {component.name} has as " \
                          f" many children ({number_of_children}) as the minimum number of required components (" \
                          f"{component.children_logic.minimum_number_of_required_component}), thus it is equivalent" \
                          f" to an And logic."
                logging.warning(message)

        elif isinstance(component.children_logic, ToleratedFault):
            if number_of_children < component.children_logic.fault_tolerance:
                message = f"{type(component.children_logic)} children logic of component {component.name} has less" \
                          f" children ({number_of_children}) than the number of tolerated faults (" \
                          f"{component.children_logic.fault_tolerance}), thus the parent component will always be" \
                          f" in a RUNNING status."
                logging.warning(message)
            elif number_of_children == component.children_logic.fault_tolerance:
                message = f"{type(component.children_logic)} children logic of component {component.name} has as " \
                          f" many children ({number_of_children}) as the number of tolerated faults (" \
                          f"{component.children_logic.fault_tolerance}), thus  the parent component will always be" \
                          f" in a RUNNING status."
                logging.warning(message)

        elif isinstance(component.children_logic, And):
            pass

        elif isinstance(component.children_logic, Oo):

            if number_of_children != component.children_logic.total_number_of_component:
                message = f"{type(component.children_logic)} children logic of component {component.name} has a " \
                          f" different number of children ({number_of_children}) than the one provided in the input" \
                          f" file ({component.children_logic.total_number_of_component})."
                logging.exception(message)
                raise ChildrenLogicSanityCheckError(message)
            elif component.children_logic.minimum_number_of_required_component > \
                    component.children_logic.total_number_of_component:
                message = f"{type(component.children_logic)} children logic of component {component.name} has less" \
                          f" children ({number_of_children}) than the minimum number required component (" \
                          f"{component.children_logic.minimum_number_of_required_component}), thus the parent " \
                          f"component can never be in a RUNNING status."
                logging.exception(message)
                raise ChildrenLogicSanityCheckError(message)
            elif number_of_children == component.children_logic.minimum_number_of_required_component:
                message = f"{type(component.children_logic)} children logic of component {component.name} has as " \
                          f" many children ({number_of_children}) than the minimum number required component (" \
                          f"{component.children_logic.minimum_number_of_required_component}), thus it is equivalent" \
                          f" to an And logic."
                logging.warning(message)

        else:
            message = f"{type(component.children_logic)} children logic of component {component.name} cannot be " \
                      f"analysed in the sanity checks, most likely it is a user defined children logic."
            logging.warning(message)

        for child in component.get_children():
            children_logic_validity(child)

def run(system_template: SystemTemplate,
        root: Component) -> bool:
    """
    Function which list all the names present in the tree (component name, failure mode name, etc) and the names
    present in the input file. A comparison is performed to do a simple, but not exhaustive, test
    """

    dict_of_lists_of_names_extracted_from_the_tree = export_content_of_tree(root)

    dict_of_lists_of_names_extracted_from_the_system_template = export_content_of_system_template(system_template)

    # logging possible differences between names extracted from the input file and
    # names extracted form the tree simulated
    matching=True
    for key in dict_of_lists_of_names_extracted_from_the_tree:
        matching = matching and _compare_and_log_diff(dict_of_lists_of_names_extracted_from_the_tree[key],
                                                      dict_of_lists_of_names_extracted_from_the_system_template[key],
                                                      key)

    # counting the number of failure modes present in the component tree
    total_number_of_failure_modes_in_the_system = 0
    for component in root.to_set():
        if isinstance(component, Basic):
            total_number_of_failure_modes_in_the_system += 1
    if total_number_of_failure_modes_in_the_system == 0:
        msg = f"No failure mode present in the system tree"
        logging.exception(msg)
        from availsim4core.src.context.system.component_tree.component_factory import EmptySystemException
        raise EmptySystemException(msg)

    children_logic_validity(root)

    return matching


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent


class PostponeCEvent(CEvent):
    """
    Class dealing with C events used to postpone some events, for example when using phases
    """
    __slots__ = 'postpone_duration', 'b_event'

    def __init__(self,
                 priority: int,
                 context: Context,
                 postpone_duration: float,
                 b_event: BEvent):
        super().__init__(priority, context)
        self.postpone_duration = postpone_duration
        self.b_event = b_event

    def __hash__(self):
        return hash((type(self), self.priority, self.postpone_duration, self.b_event))

    def __eq__(self, other):
        if not isinstance(self, PostponeCEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.postpone_duration == other.postpone_duration and \
               self.b_event == other.b_event

    def __str__(self):
        return f"PostponeCEvent:: " \
               f"priority:{self.priority} - " \
               f"postpone_duration:{self.postpone_duration} - " \
               f"b_event:{self.b_event}"

    def generate_b_events(self, absolute_simulation_time):
        return {self.b_event.postpone(self.postpone_duration)}
